import tempfile

import pandas as pd
import plotly.express as px
from flask import render_template
from flask import request
from flask import send_file

from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

FIGURE_HTML = "figure"
ERR_HTML = "<div>There is no data for this chart, please check you are looking at a valid storage service.</div>"

HEADERS = [
    fields.FIELD_UUID,
    fields.FIELD_STORAGE_LOCATION,
    fields.FIELD_AIPS,
    fields.FIELD_SIZE,
    fields.FIELD_FILE_COUNT,
]


@reporter.route("/storage_locations/", methods=["GET"])
def storage_locations():
    """Return AIPs containing file format, sorted by count and total size."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    locations_data = report_data.storage_locations(
        storage_service_id=storage_service_id, start_date=start_date, end_date=end_date
    )
    locations = locations_data.get(fields.FIELD_LOCATIONS)

    if csv:
        headers = translate_headers(HEADERS, True)

        filename = "storage_locations.csv"
        csv_data = format_size_for_csv(locations)
        return download_csv(headers, csv_data, filename)

    headers = translate_headers(HEADERS)

    return render_template(
        "report_storage_locations.html",
        storage_service_id=storage_service_id,
        storage_service_name=locations_data.get(fields.FIELD_STORAGE_NAME),
        columns=headers,
        locations=locations,
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )


def _get_line_chart_figure_html(locations, metric="aips"):
    """Return partial HTML of bar chart plot and the underlying pandas df."""
    table_dict = {"days": [], "aips": [], "size": [], "files": [], "location": []}
    for day, locations_data in locations.items():
        for location in locations_data:
            table_dict["days"].append(day)
            table_dict["aips"].append(location.get(fields.FIELD_AIPS, 0))
            # Display size in GB
            table_dict["size"].append(location.get(fields.FIELD_SIZE, 0) / 1000000)
            table_dict["files"].append(location.get(fields.FIELD_FILE_COUNT, 0))
            table_dict["location"].append(
                "{} ({})".format(location["StorageLocation"], location["UUID"])
            )

    df = pd.DataFrame(table_dict)
    fig = px.line(df, x="days", y=metric, color="location", markers=True)
    return fig.to_html(full_html=False), df


@reporter.route("/storage_locations_usage_over_time/", methods=["GET"])
def storage_location_usage_over_time():
    """Return the information needed to present a timeseries line chart."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)
    request_metric = request.args.get(request_params.METRIC, "aips")
    cumulative = parse_bool(request.args.get(request_params.CUMULATIVE), default=False)

    # Determine which metric we're looking to chart.
    metric = "aips"
    if request_metric in ("files", "size"):
        metric = request_metric

    locations_data = report_data.storage_locations_usage_over_time(
        storage_service_id=storage_service_id,
        start_date=start_date,
        end_date=end_date,
        cumulative=cumulative,
    )

    html_figure, dataframe = _get_line_chart_figure_html(
        locations_data.get(fields.FIELD_LOCATIONS_USAGE_OVER_TIME), metric=metric
    )

    if csv:
        # Use the pandas dataframe created as part of plotting our timeseries
        # to download a CSV containing the data underpinning the chart that is
        # currently displayed to the user in the UI.
        filename = "storage_locations_usage_over_time.csv"
        with tempfile.NamedTemporaryFile() as tmp:
            dataframe.to_csv(tmp.name, encoding="utf-8", index=False)
            return send_file(
                tmp.name,
                mimetype="text/csv",
                as_attachment=True,
                download_name=filename,
            )

    return render_template(
        "report_storage_locations_usage_over_time.html",
        storage_service_id=storage_service_id,
        storage_service_name=locations_data.get(fields.FIELD_STORAGE_NAME),
        plot=html_figure,
        start_date=start_date,
        end_date=get_display_end_date(end_date),
        metric=metric,
        cumulative=cumulative,
    )



import hashlib
from datetime import datetime
from datetime import timedelta


def parse_bool(value, default=True):
    if value is not None:
        value = value.lower()

        if value in ("y", "yes", "on", "1", "true", "t"):
            return True

    return False


def parse_datetime_bound(date_string, upper=False):
    """Parse date parameter string into datetime object.

    Date parameters come in from the UI or API as date strings. This
    helper converts these date strings to datetime objects for use in
    filtering by date range in Data endpoints.

    Date filtering in SQLAlchemy queries in the Data endpoint will be
    inclusive when the following rules are followed:
    - Use >= for start_date comparison
    - Use < for end_date comparison

    None or invalid start and end date values are set to datetime.min
    and datetime.max respectively.

    :param date_string: Date string. Valid format is YYYY-MM-DD (str).
    :param upper: Flag indicating if date string being parsed is
        the upper bound, with default of False (bool).

    :returns: datetime.datetime object
    """
    default_datetime = datetime.min
    if upper:
        default_datetime = datetime.max

    try:
        parsed_datetime = datetime.strptime(date_string, "%Y-%m-%d")
        if upper:
            parsed_datetime = parsed_datetime + timedelta(days=1)
        return parsed_datetime
    except (TypeError, ValueError):
        return default_datetime


def filesizeformat(value, binary=False):
    """Format the value like a 'human-readable' file size (i.e. 13 kB,
    4.1 MB, 102 Bytes, etc).  Per default decimal prefixes are used (Mega,
    Giga, etc.), if the second parameter is set to `True` the binary
    prefixes are used (Mebi, Gibi).

    Copied from jinja2 source code to ensure consistency of human-readable
    dates regardless of whether conversion happens in a view or a template.
    Source code: https://github.com/pallets/jinja/blob/
    9d4689b04d53f233b8b9ab664edb2f7430d2bbde/src/jinja2/filters.py

    Copyright 2007 Pallets
    BSD 3 license
    """
    byte_size = float(value)
    base = 1024 if binary else 1000
    prefixes = [
        ("KiB" if binary else "kB"),
        ("MiB" if binary else "MB"),
        ("GiB" if binary else "GB"),
        ("TiB" if binary else "TB"),
        ("PiB" if binary else "PB"),
        ("EiB" if binary else "EB"),
        ("ZiB" if binary else "ZB"),
        ("YiB" if binary else "YB"),
    ]

    if byte_size == 1:
        return "1 Byte"
    elif byte_size < base:
        return f"{int(byte_size)} Bytes"
    else:
        for i, prefix in enumerate(prefixes):
            unit = base ** (i + 2)

            if byte_size < unit:
                return f"{base * byte_size / unit:.1f} {prefix}"

        return f"{base * byte_size / unit:.1f} {prefix}"


def _split_ms(date_string):
    """Remove microseconds from the given date string."""
    return str(date_string).split(".")[0]


def _simplify_datetime(date_string, return_object=True):
    """Consistently return datetime string or object as required with
    microseconds striped.
    """
    DATE_FORMAT_FULL = "%Y-%m-%d %H:%M:%S"
    formatted_date = datetime.strptime(_split_ms(date_string), DATE_FORMAT_FULL)
    if return_object:
        return formatted_date
    return formatted_date.strftime(DATE_FORMAT_FULL)


def stream_write_and_hash(chunk_source, destination_path, chunk_size=65536):
    """Write bytes from a stream/iterable to disk while computing SHA256.

    The helper accepts any object that yields byte chunks: a `requests.Response`
    obtained with ``stream=True``, a file-like object exposing ``read`` or an
    arbitrary iterable of bytes. Chunks are consumed incrementally so only a
    small buffer is retained in memory. The caller is responsible for closing
    the *chunk_source* when applicable.

    :param chunk_source: Iterable, file-like object, or response providing data.
    :param destination_path: Filesystem path where the payload is written.
    :param chunk_size: Bytes to read per iteration when ``chunk_source`` supports
        size hints (defaults to 64 KiB).
    :returns: SHA256 hex digest of the written payload.
    """

    if hasattr(chunk_source, "iter_content"):
        iterator = chunk_source.iter_content(chunk_size=chunk_size)
    elif hasattr(chunk_source, "read"):

        def _reader():
            while True:
                data = chunk_source.read(chunk_size)
                if not data:
                    break
                yield data

        iterator = _reader()
    else:
        iterator = iter(chunk_source)

    sha256_hasher = hashlib.sha256()

    with open(destination_path, "wb") as destination:
        for chunk in iterator:
            if not chunk:
                continue
            destination.write(chunk)
            sha256_hasher.update(chunk)

    return sha256_hasher.hexdigest()



import pytest
from flask import current_app

EXPECTED_CSV = b"AIP UUID,AIP Name,UUID,Name,Format,Original UUID,Original Name,Original Format,Original Version,Original PUID\r\n111111111111-1111-1111-11111111,TestAIP1,555555555555-5555-5555-55555555,preservation-file-1.tif,Tagged Image File Format,333333333333-3333-3333-33333333,original-file-1.jpg,JPEG,1.01,fmt/43\r\n222222222222-2222-2222-22222222,TestAIP2,666666666666-6666-6666-66666666,preservation-file-2.tif,Tagged Image File Format,444444444444-4444-4444-44444444,original-file-2.jpg,JPEG,1.02,fmt/44\r\n"
EXPECTED_CSV_FILTERED = b"AIP UUID,AIP Name,UUID,Name,Format,Original UUID,Original Name,Original Format,Original Version,Original PUID\r\n111111111111-1111-1111-11111111,TestAIP1,555555555555-5555-5555-55555555,preservation-file-1.tif,Tagged Image File Format,333333333333-3333-3333-33333333,original-file-1.jpg,JPEG,1.01,fmt/43\r\n"


def test_preservation_derivatives(aip_contents):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/preservation_derivatives/?amss_id=1")
        assert response.status_code == 200


@pytest.mark.parametrize(
    "aip_uuid,expected_csv_contents",
    [
        # No AIP UUID returns all results.
        ("", EXPECTED_CSV),
        # Specifying an AIP UUID filters results to that AIP.
        ("111111111111-1111-1111-11111111", EXPECTED_CSV_FILTERED),
    ],
)
def test_preservation_derivatives_csv(
    preservation_derivatives, aip_uuid, expected_csv_contents
):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/preservation_derivatives/?amss_id=1&aip_uuid={aip_uuid}&csv=True"
        )
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=preservation_derivatives.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == expected_csv_contents



import pytest
from flask import current_app
from flask import url_for

from AIPscan import typesense_test_helpers

EXPECTED_CSV_CONTENTS = b"Format,Count,Size,Size (bytes)\r\nJPEG,2,3.0 kB,3000\r\nISO Disk Image File,1,0 Bytes,0\r\n"


@pytest.mark.parametrize(
    "view, url_params",
    [
        ("reporter.report_formats_count", {"amss_id": "1"}),
        (
            "reporter.chart_formats_count",
            {
                "start_date": "2019-01-30",
                "end_date": "2024-02-26",
                "amss_id": 1,
                "storage_location": "",
            },
        ),
        (
            "reporter.plot_formats_count",
            {
                "start_date": "2019-01-30",
                "end_date": "2024-02-26",
                "amss_id": "1",
                "storage_location": "",
            },
        ),
    ],
)
def test_formats_count_reports(app_with_populated_format_versions, view, url_params):
    """Test that report template renders."""
    with current_app.test_request_context():
        url_path_and_params = url_for(view, **url_params)

    with current_app.test_client() as test_client:
        response = test_client.get(url_path_and_params)
        assert response.status_code == 200


@pytest.mark.parametrize(
    "view, url_params",
    [
        ("reporter.report_formats_count", {"amss_id": 1}),
        (
            "reporter.chart_formats_count",
            {
                "start_date": "2019-01-30",
                "end_date": "2024-02-26",
                "amss_id": "1",
                "storage_location": "",
            },
        ),
        (
            "reporter.plot_formats_count",
            {
                "start_date": "2019-01-30",
                "end_date": "2024-02-26",
                "amss_id": "1",
                "storage_location": "",
            },
        ),
    ],
)
def test_formats_count_reports_using_typesense(
    app_with_populated_format_versions, enable_typesense, mocker, view, url_params
):
    """Test that report template renders."""
    with current_app.test_request_context():
        url_path_and_params = url_for(view, **url_params)

    with current_app.test_client() as test_client:
        typesense_test_helpers.fake_collection_format_counts(mocker)

        response = test_client.get(url_path_and_params)
        assert response.status_code == 200


def test_formats_count_csv(app_with_populated_format_versions):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/report_formats_count/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=file_formats.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS



import datetime
import uuid

import pytest

from AIPscan import db
from AIPscan import typesense_helpers
from AIPscan import typesense_test_helpers
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.Data.tests import MOCK_AIP
from AIPscan.Data.tests import MOCK_AIP_NAME
from AIPscan.Data.tests import MOCK_AIP_UUID
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_ID
from AIPscan.helpers import parse_datetime_bound
from AIPscan.models import File
from AIPscan.models import FileType

TEST_FILES = [
    File(
        uuid=uuid.uuid4(),
        name="test.csv",
        size=1234567,
        aip_id=1,
        file_type=FileType.original,
        file_format="Comma Separated Values",
        filepath="/path/to/file.csv",
        date_created=datetime.datetime.now(),
        checksum_type="md5",
        checksum_value="fakemd5",
    ),
    File(
        uuid=uuid.uuid4(),
        name="test.txt",
        size=12345,
        aip_id=2,
        file_type=FileType.original,
        file_format="Plain Text File",
        puid="x-fmt/111",
        filepath="/path/to/file.txt",
        date_created=datetime.datetime.now(),
        checksum_type="md5",
        checksum_value="anotherfakemd5",
    ),
    File(
        uuid=uuid.uuid4(),
        name="test.pdf",
        size=12345678,
        aip_id=1,
        file_type=FileType.preservation,
        file_format="Acrobat PDF/A - Portable Document Format",
        format_version="1b",
        filepath="/path/to/test.pdf",
        date_created=datetime.datetime.now(),
        checksum_type="md5",
        checksum_value="yetanotherfakemd5",
        original_file_id=1,
    ),
]


@pytest.mark.parametrize(
    "storage_location_id, storage_location_description, start_date, end_date, file_count, largest_file_size, second_largest_file_size",
    [
        # Test all AIPs.
        (None, None, "2020-01-01", "2022-12-31", 7, 2500, 2500),
        # Test filtering by date.
        (None, None, "2020-01-01", "2020-06-01", 4, 1000, 300),
        # Test filtering by storage location.
        (1, "AIP Store Location 1", "2020-01-01", "2022-12-31", 4, 1000, 300),
        (2, "AIP Store Location 2", "2020-01-01", "2022-12-31", 3, 2500, 2500),
    ],
)
def test_largest_files(
    storage_locations,
    storage_location_id,
    storage_location_description,
    start_date,
    end_date,
    file_count,
    largest_file_size,
    second_largest_file_size,
):
    """Test that files and order returned by report_data.largest_files match expectations."""
    report = report_data.largest_files(
        storage_service_id=1,
        start_date=parse_datetime_bound(start_date),
        end_date=parse_datetime_bound(end_date, upper=True),
        storage_location_id=storage_location_id,
    )
    report_files = report[fields.FIELD_FILES]
    assert report[fields.FIELD_STORAGE_NAME] == "test storage service"
    assert report[fields.FIELD_STORAGE_LOCATION] == storage_location_description

    assert len(report_files) == file_count
    assert report_files[0][fields.FIELD_SIZE] == largest_file_size
    assert report_files[1][fields.FIELD_SIZE] == second_largest_file_size


@pytest.mark.parametrize(
    "test_file, has_format_version, has_puid",
    [
        (TEST_FILES[0], False, False),
        (TEST_FILES[1], False, True),
        (TEST_FILES[2], True, False),
    ],
)
def test_largest_files_elements(
    app_instance, mocker, test_file, has_format_version, has_puid
):
    """Test that returned file data matches expected values."""
    mock_query = mocker.patch("AIPscan.Data.report_data._largest_files_query")
    mock_query.return_value = [test_file]

    mock_get_ss_name = mocker.patch("AIPscan.Data._get_storage_service")
    mock_get_ss_name.return_value = MOCK_STORAGE_SERVICE

    mock_get_aip = mocker.patch("AIPscan.Data.report_data.db.session.get")
    mock_get_aip.return_value = MOCK_AIP

    report = report_data.largest_files(
        MOCK_STORAGE_SERVICE_ID,
        start_date=parse_datetime_bound("2000-01-01"),
        end_date=parse_datetime_bound("2022-12-31", upper=True),
    )
    report_file = report[fields.FIELD_FILES][0]

    # Required elements
    assert test_file.name == report_file.get(fields.FIELD_NAME)
    assert test_file.file_format == report_file.get(fields.FIELD_FORMAT)

    # Optional elements
    if has_format_version:
        assert test_file.format_version == report_file.get(fields.FIELD_VERSION)
    else:
        assert report_file.get(fields.FIELD_VERSION) is None

    if has_puid:
        assert test_file.puid == report_file.get(fields.FIELD_PUID)
    else:
        assert report_file.get(fields.FIELD_PUID) is None

    # AIP information
    assert report_file.get(fields.FIELD_AIP_NAME) == MOCK_AIP_NAME
    assert report_file.get(fields.FIELD_AIP_UUID) == MOCK_AIP_UUID


def test_largest_files_typesense(app_with_populated_files, enable_typesense, mocker):
    doc = typesense_helpers.model_instance_to_document(File, db.session.get(File, 1))
    doc["transfer_name"] = "Test AIP"
    doc["aip_uuid"] = "111111111111-1111-1111-11111111"

    fake_results = {"hits": [{"document": doc}]}

    typesense_test_helpers.fake_collection(mocker, fake_results)

    expected_result = {
        "Files": [
            {
                "AIPName": "Test AIP",
                "AIPUUID": "111111111111-1111-1111-11111111",
                "ID": "1",
                "UUID": "222222222222-2222-2222-22222222",
                "Name": "file_name.ext",
                "Size": 1000,
                "FileType": "original",
                "Format": "Tagged Image File Format",
                "Version": "0.0.0",
                "PUID": "fmt/353",
            }
        ],
        "StorageName": "test storage service",
        "StorageLocation": "test storage location",
    }

    report = report_data_typesense.largest_files(
        storage_service_id=1,
        start_date=parse_datetime_bound("2019-01-01"),
        end_date=parse_datetime_bound("2019-10-10"),
        storage_location_id=1,
    )
    assert report == expected_result



"""Report formats count consists of the tabular report, plot, and
chart which describe the file formats present across the AIPs in a
storage service with AIPs filtered by date range.
"""

from collections import Counter
from datetime import datetime
from datetime import timedelta

from flask import render_template
from flask import request

from AIPscan import typesense_helpers as ts_helpers
from AIPscan.Data import fields
from AIPscan.Data import get_storage_location_description
from AIPscan.Data import get_storage_service_name
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.helpers import filesizeformat
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.models import AIP
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

HEADERS = [fields.FIELD_FORMAT, fields.FIELD_COUNT, fields.FIELD_SIZE]


@reporter.route("/report_formats_count/", methods=["GET"])
def report_formats_count():
    """Report (tabular) on all file formats and their counts and size on
    disk across all AIPs in the storage service.
    """
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    if ts_helpers.typesense_enabled():
        formats_data = report_data_typesense.formats_count(
            storage_service_id, storage_location_id, start_date, end_date
        )
    else:
        formats_data = report_data.formats_count(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
        )

    formats = formats_data.get(fields.FIELD_FORMATS)

    if csv:
        headers = translate_headers(HEADERS, True)

        filename = "file_formats.csv"
        csv_data = format_size_for_csv(formats)
        return download_csv(headers, csv_data, filename)

    headers = translate_headers(HEADERS)

    return render_template(
        "report_formats_count.html",
        storage_service_id=storage_service_id,
        storage_service_name=formats_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=formats_data.get(fields.FIELD_STORAGE_LOCATION),
        columns=headers,
        formats=formats,
        total_file_count=sum(format_.get(fields.FIELD_COUNT, 0) for format_ in formats),
        total_size=sum(format_.get(fields.FIELD_SIZE, 0) for format_ in formats),
        formats_count=len(formats),
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )


@reporter.route("/chart_formats_count/", methods=["GET"])
def chart_formats_count():
    """Report (pie chart) on all file formats and their counts and size
    on disk across all AIPs in the storage service."""
    start_date = request.args.get(request_params.START_DATE)
    end_date = request.args.get(request_params.END_DATE)
    # make date range inclusive
    start = datetime.strptime(start_date, "%Y-%m-%d")
    end = datetime.strptime(end_date, "%Y-%m-%d")
    day_before = start - timedelta(days=1)
    day_after = end + timedelta(days=1)

    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)

    if ts_helpers.typesense_enabled():
        report = report_data_typesense.formats_count(
            storage_service_id, storage_location_id, day_before, day_after, False
        )

        format_counts = dict(
            zip(
                [item["Format"] for item in report[fields.FIELD_FORMATS]],
                [item["Count"] for item in report[fields.FIELD_FORMATS]],
                strict=False,
            )
        )
    else:
        aips = AIP.query.filter_by(storage_service_id=storage_service_id)
        if storage_location_id:
            aips = aips.filter_by(storage_location_id=storage_location_id)
        aips = aips.all()

        format_labels = []
        format_counts = []

        for aip in aips:
            original_files = File.query.filter_by(
                aip_id=aip.id, file_type=FileType.original
            )
            for original in original_files:
                if aip.create_date < day_before:
                    continue
                elif aip.create_date > day_after:
                    continue
                else:
                    format_labels.append(original.file_format)

        format_counts = Counter(format_labels)

    labels = list(format_counts.keys())
    labels.sort()
    values = list(format_counts.values())

    originals_count = sum(values)
    different_formats = len(format_counts.keys())

    return render_template(
        "chart_formats_count.html",
        startdate=start_date,
        enddate=end_date,
        storage_service_name=get_storage_service_name(storage_service_id),
        storage_location_description=get_storage_location_description(
            storage_location_id
        ),
        labels=labels,
        values=values,
        originalsCount=originals_count,
        differentFormats=different_formats,
    )


@reporter.route("/plot_formats_count/", methods=["GET"])
def plot_formats_count():
    """Report (scatter) on all file formats and their counts and size on
    disk across all AIPs in the storage service.
    """
    start_date = request.args.get(request_params.START_DATE)
    end_date = request.args.get(request_params.END_DATE)
    # make date range inclusive
    start = datetime.strptime(start_date, "%Y-%m-%d")
    end = datetime.strptime(end_date, "%Y-%m-%d")
    day_before = start - timedelta(days=1)
    day_after = end + timedelta(days=1)

    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)

    if ts_helpers.typesense_enabled():
        report = report_data_typesense.formats_count(
            storage_service_id, storage_location_id, day_before, day_after
        )

        format_count = {}
        originals_count = 0

        for file_format in report[fields.FIELD_FORMATS]:
            format_count[file_format["Format"]] = {
                "count": file_format["Count"],
                "size": file_format["Size"],
            }

            originals_count += file_format["Count"]
    else:
        aips = AIP.query.filter_by(storage_service_id=storage_service_id)
        if storage_location_id:
            aips = aips.filter_by(storage_location_id=storage_location_id)
        aips = aips.all()

        format_count = {}
        originals_count = 0

        for aip in aips:
            original_files = File.query.filter_by(
                aip_id=aip.id, file_type=FileType.original
            )
            for original in original_files:
                if aip.create_date < day_before:
                    continue
                elif aip.create_date > day_after:
                    continue
                else:
                    originals_count += 1
                    file_format = original.file_format
                    size = original.size

                    if file_format in format_count:
                        format_count[file_format]["count"] += 1
                        format_count[file_format]["size"] += size
                    else:
                        format_count.update({file_format: {"count": 1, "size": size}})

    total_size = 0
    x_axis = []
    y_axis = []
    file_format = []
    human_size = []

    for _, value in format_count.items():
        y_axis.append(value["count"])
        size = value["size"]
        if size is None:
            size = 0
        x_axis.append(size)
        total_size += size
        human_size.append(filesizeformat(size))

    file_format = list(format_count.keys())
    different_formats = len(format_count.keys())
    total_human_size = filesizeformat(total_size)

    return render_template(
        "plot_formats_count.html",
        startdate=start_date,
        enddate=end_date,
        storage_service_name=get_storage_service_name(storage_service_id),
        storage_location_description=get_storage_location_description(
            storage_location_id
        ),
        originalsCount=originals_count,
        formatCount=format_count,
        differentFormats=different_formats,
        totalHumanSize=total_human_size,
        x_axis=x_axis,
        y_axis=y_axis,
        format=file_format,
        humansize=human_size,
    )



FAKE_RESULTS_FORMAT_COUNTS = {
    "facet_counts": [
        {"field_name": "file_format", "counts": [{"value": "wav", "count": 10}]}
    ]
}


FAKE_MULTI_SEARCH_RESULTS_FORMAT_COUNTS = {
    "results": [
        {
            "hits": [{"document": {"file_format": "wav"}}],
            "facet_counts": [{"field_name": "size", "stats": {"sum": 999}}],
        }
    ]
}


def fake_collection_format_counts(mocker):
    fake_collection = FakeCollection(FAKE_RESULTS_FORMAT_COUNTS)

    query_collection = mocker.patch("typesense.collections.Collections.__getitem__")
    query_collection.return_value = fake_collection

    query_multi = mocker.patch("typesense.multi_search.MultiSearch.perform")
    query_multi.return_value = FAKE_MULTI_SEARCH_RESULTS_FORMAT_COUNTS

    return fake_collection


def fake_collection(mocker, fake_results):
    fake_collection = FakeCollection(fake_results)

    query = mocker.patch("typesense.collections.Collections.__getitem__")
    query.return_value = fake_collection


class FakeDocuments:
    def __init__(self, fake_results):
        self.fake_results = fake_results

    def search(self, search_parameters):
        return self.fake_results

    def _import(self, search_parameters):
        return {}


class FakeCollection:
    def __init__(self, fake_results=None):
        self.documents = FakeDocuments(fake_results)

    def delete(self):
        return



"""This module defines shared AIPscan pytest fixtures."""

import uuid
from datetime import datetime

import pytest
from flask import current_app

from AIPscan import create_app
from AIPscan import db
from AIPscan import test_helpers
from AIPscan.models import FileType

AIP_DATE_FORMAT = "%Y-%m-%d"

TIFF_FILE_FORMAT = "Tagged Image File Format"
TIFF_PUID = "fmt/353"

JPEG_FILE_FORMAT = "JPEG"

JPEG_1_01_FORMAT_VERSION = "1.01"
JPEG_1_01_PUID = "fmt/43"

JPEG_1_02_FORMAT_VERSION = "1.02"
JPEG_1_02_PUID = "fmt/44"

ORIGINAL_FILE_SIZE = 1000
PRESERVATION_FILE_SIZE = 2000

AIP_1_CREATION_DATE = "2020-01-01"
AIP_2_CREATION_DATE = "2020-06-01"
AIP_3_CREATION_DATE = "2021-05-31"

AIP_1_NAME = "TestAIP1"
AIP_2_NAME = "TestAIP2"
AIP_3_NAME = "TestAIP3"

AIP_1_UUID = "111111111111-1111-1111-11111111"
AIP_2_UUID = "222222222222-2222-2222-22222222"
AIP_3_UUID = "333333333333-3333-3333-33333333"

ORIGINAL_FILE_1_NAME = "original-file-1.jpg"
ORIGINAL_FILE_1_UUID = "333333333333-3333-3333-33333333"

ORIGINAL_FILE_2_NAME = "original-file-2.jpg"
ORIGINAL_FILE_2_UUID = "444444444444-4444-4444-44444444"

ORIGIN_PIPELINE_UUID = "777777777777-7777-7777-77777777"
ORIGIN_PIPELINE = f"/api/v2/pipeline/{ORIGIN_PIPELINE_UUID}/"

PRESERVATION_FILE_1_NAME = "preservation-file-1.tif"
PRESERVATION_FILE_2_NAME = "preservation-file-2.tif"

PRESERVATION_FILE_1_UUID = "555555555555-5555-5555-55555555"
PRESERVATION_FILE_2_UUID = "666666666666-6666-6666-66666666"

PRESERVATION_FORMAT = "Very special preservation format"
PRESERVATION_PUID = "fmt/000"

PUID_1 = "fmt/43"
PUID_2 = "fmt/61"
PUID_3 = "x-fmt/111"

STORAGE_LOCATION_1_UUID = "2bbcea40-eb4d-4076-a81d-1ab046e34f6a"
STORAGE_LOCATION_1_CURRENT_LOCATION = f"/api/v2/location/{STORAGE_LOCATION_1_UUID}/"
STORAGE_LOCATION_1_DESCRIPTION = "AIP Store Location 1"

STORAGE_LOCATION_2_UUID = "e69beb57-0e32-4c45-8db7-9b7723724a05"
STORAGE_LOCATION_2_CURRENT_LOCATION = f"/api/v2/location/{STORAGE_LOCATION_2_UUID}/"
STORAGE_LOCATION_2_DESCRIPTION = "AIP Store Location 2"

STORAGE_SERVICE_NAME = "Test Storage Service"


@pytest.fixture(scope="session")
def app_setup():
    """Session-wide test `Flask` application setup.

    Most tests should use the `app_instance` fixture instead, which
    ensures a clean database state for each test.
    """
    app = create_app("test")
    with app.app_context():
        db.create_all()
        yield app
        db.session.remove()
        db.drop_all()
        db.engine.dispose()


@pytest.fixture(scope="function")
def app_instance(app_setup):
    """Function-scoped test `Flask` application instance.

    It ensures a clean database state for each test by truncating all tables
    before yielding the application instance.
    """
    with app_setup.app_context():
        db.session.execute(db.text("SET FOREIGN_KEY_CHECKS=0;"))
        for table in db.metadata.tables:
            db.session.execute(db.text(f"TRUNCATE TABLE `{table}`;"))
        db.session.execute(db.text("SET FOREIGN_KEY_CHECKS=1;"))
        db.session.commit()
        yield app_setup
        db.session.remove()


def _datetime_obj_from_string(date_string):
    """Return a datetime object from a given string to aid in the
    assignment of consistent and reliable test dates.
    """
    return datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S")


INGEST_EVENT_CREATION_TIME = _datetime_obj_from_string("2020-12-02 10:00:00")
AIP_CREATION_TIME = _datetime_obj_from_string("2020-12-02 10:30:32")


@pytest.fixture
def app_with_populated_files(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the Data.aips_by_file_format and Data.aips_by_puid endpoints.
    """
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    aip = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        create_date=AIP_CREATION_TIME,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    original_file = test_helpers.create_test_file(
        file_type=FileType.original,
        size=ORIGINAL_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        aip_id=aip.id,
    )

    _ = test_helpers.create_test_file(
        file_type=FileType.preservation,
        size=PRESERVATION_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        aip_id=aip.id,
    )

    user_agent = test_helpers.create_test_agent(storage_service_id=storage_service.id)

    event = test_helpers.create_test_event(
        event_type="ingestion",
        date=INGEST_EVENT_CREATION_TIME,
        file_id=original_file.id,
    )

    _ = test_helpers.create_test_event_agent(event_id=event.id, agent_id=user_agent.id)


@pytest.fixture
def app_with_populated_files_no_ingestion_event(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the Data.aips_by_file_format and Data.aips_by_puid endpoints.
    """
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    aip = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        create_date=AIP_CREATION_TIME,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    test_helpers.create_test_file(
        file_type=FileType.original,
        size=ORIGINAL_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        aip_id=aip.id,
    )

    _ = test_helpers.create_test_file(
        file_type=FileType.preservation,
        size=PRESERVATION_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        aip_id=aip.id,
    )


@pytest.fixture
def app_with_populated_format_versions(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the Data.format_versions_count endpoint.
    """
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    aip1 = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        create_date=datetime.strptime(AIP_1_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    aip2 = test_helpers.create_test_aip(
        uuid=AIP_2_UUID,
        create_date=datetime.strptime(AIP_2_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    _ = test_helpers.create_test_file(
        size=ORIGINAL_FILE_SIZE,
        uuid=ORIGINAL_FILE_1_UUID,
        name="original.jpg",
        puid=JPEG_1_01_PUID,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_01_FORMAT_VERSION,
        aip_id=aip1.id,
    )

    _ = test_helpers.create_test_file(
        size=PRESERVATION_FILE_SIZE,
        uuid=PRESERVATION_FILE_1_UUID,
        name="preservation.jpg",
        puid=JPEG_1_02_PUID,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_02_FORMAT_VERSION,
        aip_id=aip2.id,
    )

    _ = test_helpers.create_test_file(
        size=None,
        uuid=ORIGINAL_FILE_2_UUID,
        name="original.iso",
        puid="fmt/468",
        file_format="ISO Disk Image File",
        format_version=None,
        aip_id=aip2.id,
    )


@pytest.fixture
def preservation_derivatives(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the report_data.preservation_derivatives endpoint.
    """
    storage_service = test_helpers.create_test_storage_service(
        name=STORAGE_SERVICE_NAME
    )
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    aip1 = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        transfer_name=AIP_1_NAME,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )
    aip2 = test_helpers.create_test_aip(
        uuid=AIP_2_UUID,
        transfer_name=AIP_2_NAME,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    original_file1 = test_helpers.create_test_file(
        file_type=FileType.original,
        name=ORIGINAL_FILE_1_NAME,
        uuid=ORIGINAL_FILE_1_UUID,
        size=ORIGINAL_FILE_SIZE,
        puid=JPEG_1_01_PUID,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_01_FORMAT_VERSION,
        aip_id=aip1.id,
    )
    original_file2 = test_helpers.create_test_file(
        file_type=FileType.original,
        name=ORIGINAL_FILE_2_NAME,
        uuid=ORIGINAL_FILE_2_UUID,
        size=ORIGINAL_FILE_SIZE,
        puid=JPEG_1_02_PUID,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_02_FORMAT_VERSION,
        aip_id=aip2.id,
    )

    _ = test_helpers.create_test_file(
        file_type=FileType.preservation,
        name=PRESERVATION_FILE_1_NAME,
        uuid=PRESERVATION_FILE_1_UUID,
        size=PRESERVATION_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        original_file_id=original_file1.id,
        aip_id=aip1.id,
    )
    _ = test_helpers.create_test_file(
        file_type=FileType.preservation,
        name=PRESERVATION_FILE_2_NAME,
        uuid=PRESERVATION_FILE_2_UUID,
        size=PRESERVATION_FILE_SIZE,
        puid=TIFF_PUID,
        file_format=TIFF_FILE_FORMAT,
        original_file_id=original_file2.id,
        aip_id=aip2.id,
    )


@pytest.fixture
def aip_contents(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the Data.aips_by_file_format and Data.aips_by_puid endpoints.
    """
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    aip1 = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        create_date=datetime.strptime(AIP_1_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    aip2 = test_helpers.create_test_aip(
        uuid=AIP_2_UUID,
        create_date=datetime.strptime(AIP_2_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
    )

    # Create files associated with AIP 1 :
    # - One file with PUID_1
    # - One file with PUID_2
    _ = test_helpers.create_test_file(puid=PUID_1, aip_id=aip1.id)
    _ = test_helpers.create_test_file(puid=PUID_2, aip_id=aip1.id)

    # Create files associated with AIP 2:
    # - Two files with PUID_2
    # - Three files with PUID_3
    _ = test_helpers.create_test_file(puid=PUID_2, aip_id=aip2.id)
    _ = test_helpers.create_test_file(puid=PUID_2, aip_id=aip2.id)
    _ = test_helpers.create_test_file(puid=PUID_3, aip_id=aip2.id)
    _ = test_helpers.create_test_file(puid=PUID_3, aip_id=aip2.id)
    _ = test_helpers.create_test_file(puid=PUID_3, aip_id=aip2.id)


@pytest.fixture
def largest_aips(app_instance):
    """Fixture with pre-populated data.

    This fixture is used to create expected state which is then used to
    test the Data.report_data.largest_aips endpoint.
    """
    storage_service = test_helpers.create_test_storage_service()
    storage_location_1 = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id, description="storage location 1"
    )
    storage_location_2 = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id,
        description="storage location 2",
        current_location="/api/v2/location/test-location-2/",
    )
    _ = test_helpers.create_test_pipeline(storage_service_id=storage_service.id)
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    # Create two AIPs in first storage location.
    test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        create_date=datetime.strptime(AIP_1_CREATION_DATE, AIP_DATE_FORMAT),
        size=10000,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location_1.id,
        fetch_job_id=fetch_job.id,
    )

    test_helpers.create_test_aip(
        uuid=AIP_2_UUID,
        create_date=datetime.strptime(AIP_2_CREATION_DATE, AIP_DATE_FORMAT),
        size=250,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location_1.id,
        fetch_job_id=fetch_job.id,
    )

    # Create three AIPS in second storage location.
    test_helpers.create_test_aip(
        uuid=AIP_3_UUID,
        create_date=datetime.strptime(AIP_3_CREATION_DATE, AIP_DATE_FORMAT),
        size=500,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location_2.id,
        fetch_job_id=fetch_job.id,
    )

    test_helpers.create_test_aip(
        uuid=str(uuid.uuid4()),
        create_date=datetime.strptime(AIP_3_CREATION_DATE, AIP_DATE_FORMAT),
        size=25,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location_2.id,
        fetch_job_id=fetch_job.id,
    )

    test_helpers.create_test_aip(
        uuid=str(uuid.uuid4()),
        create_date=datetime.strptime(AIP_3_CREATION_DATE, AIP_DATE_FORMAT),
        size=750,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location_2.id,
        fetch_job_id=fetch_job.id,
    )


@pytest.fixture
def storage_locations(app_instance):
    """Fixture with pre-populated Storage locations."""
    storage_service = test_helpers.create_test_storage_service()
    storage_location1 = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id,
        current_location=STORAGE_LOCATION_1_CURRENT_LOCATION,
        description=STORAGE_LOCATION_1_DESCRIPTION,
    )
    storage_location2 = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id,
        current_location=STORAGE_LOCATION_2_CURRENT_LOCATION,
        description=STORAGE_LOCATION_2_DESCRIPTION,
    )
    _ = test_helpers.create_test_pipeline(
        origin_pipeline=ORIGIN_PIPELINE, storage_service_id=storage_service.id
    )
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    # Create two AIPs associated with Storage Location 1.
    aip1 = test_helpers.create_test_aip(
        uuid=AIP_1_UUID,
        transfer_name=AIP_1_NAME,
        create_date=datetime.strptime(AIP_1_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location1.id,
        fetch_job_id=fetch_job.id,
    )
    aip2 = test_helpers.create_test_aip(
        uuid=AIP_2_UUID,
        transfer_name=AIP_2_NAME,
        create_date=datetime.strptime(AIP_2_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location1.id,
        fetch_job_id=fetch_job.id,
    )

    # Create one AIP associated with Storage Location 2.
    aip3 = test_helpers.create_test_aip(
        uuid=AIP_3_UUID,
        transfer_name=AIP_3_NAME,
        create_date=datetime.strptime(AIP_3_CREATION_DATE, AIP_DATE_FORMAT),
        storage_service_id=storage_service.id,
        storage_location_id=storage_location2.id,
        fetch_job_id=fetch_job.id,
    )

    # Create files associated with AIP 1, each 300 bytes, plus empty file:
    _ = test_helpers.create_test_file(
        size=0, file_type=FileType.original, file_format="txt", aip_id=aip1.id
    )
    _ = test_helpers.create_test_file(
        size=300,
        file_type=FileType.preservation,
        file_format=TIFF_FILE_FORMAT,
        puid=TIFF_PUID,
        aip_id=aip1.id,
    )
    _ = test_helpers.create_test_file(
        size=300,
        file_type=FileType.original,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_01_FORMAT_VERSION,
        puid=JPEG_1_01_PUID,
        aip_id=aip1.id,
    )

    # Create a file associated with AIP 2 of 1000 bytes:
    _ = test_helpers.create_test_file(
        size=1000,
        file_type=FileType.original,
        file_format=JPEG_FILE_FORMAT,
        format_version=JPEG_1_02_FORMAT_VERSION,
        puid=JPEG_1_02_PUID,
        aip_id=aip2.id,
    )

    # Create files associated with AIP 3, each 2500 bytes, plus empty file:
    _ = test_helpers.create_test_file(
        size=2500,
        file_type=FileType.preservation,
        file_format=PRESERVATION_FORMAT,
        puid=PRESERVATION_PUID,
        aip_id=aip3.id,
    )
    _ = test_helpers.create_test_file(
        size=2500, file_type=FileType.original, aip_id=aip3.id
    )
    _ = test_helpers.create_test_file(
        size=0,
        file_type=FileType.original,
        file_format="yet another format",
        aip_id=aip3.id,
    )


@pytest.fixture
def enable_typesense(app_instance):
    with app_instance.app_context():
        previous_api_key = current_app.config.get("TYPESENSE_API_KEY")
        current_app.config["TYPESENSE_API_KEY"] = previous_api_key or "x0x0x0x0x0"
    yield
    with app_instance.app_context():
        current_app.config["TYPESENSE_API_KEY"] = previous_api_key



import pytest

from AIPscan.Aggregator import types


def test_storage_service_package_init():
    """Test that the type constructor works as expected."""
    package_1 = types.StorageServicePackage()
    package_2 = types.StorageServicePackage()
    assert package_1 == package_2
    assert package_1.aip is False
    assert package_1.dip is False
    assert package_1.sip is False
    assert package_1.replica is False
    assert package_1.deleted is False


def test_storage_service_package_eq():
    """Provide some other equality tests for the type."""
    package_1 = types.StorageServicePackage(
        deleted=True, replica=False, uuid="1", current_path="2"
    )
    package_2 = types.StorageServicePackage(
        deleted=True, replica=False, uuid="1", current_path="2"
    )
    assert package_1 == package_2
    assert package_1.replica is False
    assert package_1.deleted is True
    package_3 = types.StorageServicePackage(
        deleted=1, replica=2, aip=3, dip=4, sip=5, uuid="6", current_path="7"
    )
    package_4 = types.StorageServicePackage(
        deleted=1, replica=2, aip=3, dip=4, sip=5, uuid="6", current_path="7"
    )
    assert package_3 == package_4
    assert package_3 != package_1


def test_package_ness():
    """Given specific package types check that they evaluate to true
    when queried.
    """
    package = types.StorageServicePackage(aip=True)
    assert package.is_aip()
    assert package.is_undeleted_aip()
    assert not package.is_dip()
    assert not package.is_sip()
    assert not package.is_deleted()
    assert not package.is_replica()

    package = types.StorageServicePackage(dip=True)
    assert package.is_dip()
    assert not package.is_aip()
    assert not package.is_undeleted_aip()
    assert not package.is_sip()
    assert not package.is_deleted()
    assert not package.is_replica()

    package = types.StorageServicePackage(sip=True)
    assert package.is_sip()
    assert not package.is_dip()
    assert not package.is_aip()
    assert not package.is_undeleted_aip()
    assert not package.is_deleted()
    assert not package.is_replica()

    package = types.StorageServicePackage(deleted=True)
    assert package.is_deleted()
    assert not package.is_replica()
    assert not package.is_aip()
    assert not package.is_undeleted_aip()
    assert not package.is_dip()
    assert not package.is_aip()

    package = types.StorageServicePackage(aip=True, deleted=True)
    assert package.is_deleted()
    assert not package.is_replica()
    assert not package.is_aip()
    assert not package.is_undeleted_aip()
    assert not package.is_dip()
    assert not package.is_aip()

    package = types.StorageServicePackage(aip=True, replica=True)
    assert package.is_replica()
    assert not package.is_deleted()
    assert not package.is_aip()
    assert not package.is_undeleted_aip()
    assert not package.is_dip()
    assert not package.is_aip()


@pytest.mark.parametrize(
    "package, result",
    [
        (
            types.StorageServicePackage(
                uuid="91940daf-5c33-4670-bfc8-9108f32bca7b",
                current_path="9194/0daf/5c33/4670/bfc8/9108/f32b/ca7b/repl-91940daf-5c33-4670-bfc8-9108f32bca7b.7z",
            ),
            "repl-91940daf-5c33-4670-bfc8-9108f32bca7b/data/METS.91940daf-5c33-4670-bfc8-9108f32bca7b.xml",
        ),
        (
            types.StorageServicePackage(
                uuid="902192d9-6232-4d5f-b6c4-dfa7b8733960",
                current_path="9021/92d9/6232/4d5f/b6c4/dfa7/b873/3960/repl-902192d9-6232-4d5f-b6c4-dfa7b8733960.7z",
            ),
            "repl-902192d9-6232-4d5f-b6c4-dfa7b8733960/data/METS.902192d9-6232-4d5f-b6c4-dfa7b8733960.xml",
        ),
        (
            types.StorageServicePackage(
                deleted=True,
                uuid="594a03f3-d8eb-4c83-affd-aefaf75d53cc",
                current_path="594a/03f3/d8eb/4c83/affd/aefa/f75d/53cc/delete_me-594a03f3-d8eb-4c83-affd-aefaf75d53cc.7z",
            ),
            None,
        ),
        (
            types.StorageServicePackage(
                uuid="e422c724-834c-4164-a7be-4c881043a531",
                current_path="e422/c724/834c/4164/a7be/4c88/1043/a531/normalize-e422c724-834c-4164-a7be-4c881043a531.7z",
            ),
            "normalize-e422c724-834c-4164-a7be-4c881043a531/data/METS.e422c724-834c-4164-a7be-4c881043a531.xml",
        ),
        (
            types.StorageServicePackage(
                uuid="583c009c-4255-44b9-8868-f57e4bade93c",
                current_path="583c/009c/4255/44b9/8868/f57e/4bad/e93c/normal_aip-583c009c-4255-44b9-8868-f57e4bade93c.7z",
            ),
            "normal_aip-583c009c-4255-44b9-8868-f57e4bade93c/data/METS.583c009c-4255-44b9-8868-f57e4bade93c.xml",
        ),
        (
            types.StorageServicePackage(
                uuid="846fca2b-0919-4673-804c-0f626a30cabd",
                current_path="846f/ca2b/0919/4673/804c/0f62/6a30/cabd/uncomp-846fca2b-0919-4673-804c-0f626a30cabd",
            ),
            "uncomp-846fca2b-0919-4673-804c-0f626a30cabd/data/METS.846fca2b-0919-4673-804c-0f626a30cabd.xml",
        ),
        (
            types.StorageServicePackage(
                dip=True,
                uuid="59f70134-eeca-4886-888e-b2013a08571e",
                current_path="1555/04f5/98dd/48d1/9e97/83bc/f0a5/efa2/normcore-59f70134-eeca-4886-888e-b2013a08571e",
            ),
            None,
        ),
        (
            types.StorageServicePackage(
                uuid="59f70134-eeca-4886-888e-b2013a08571e",
                current_path="59f7/0134/eeca/4886/888e/b201/3a08/571e/normcore-59f70134-eeca-4886-888e-b2013a08571e",
            ),
            "normcore-59f70134-eeca-4886-888e-b2013a08571e/data/METS.59f70134-eeca-4886-888e-b2013a08571e.xml",
        ),
        (
            types.StorageServicePackage(
                sip=True,
                uuid="fbdcd607-270e-4dff-9a01-d11b7c2a0200",
                current_path="originals/backlog-fbdcd607-270e-4dff-9a01-d11b7c2a0200",
            ),
            None,
        ),
    ],
)
def test_get_relative_path(package, result):
    """Given the current_path check our ability to construct a relative
    path within a METS file.
    """
    if package.replica or package.sip or package.deleted or package.dip:
        with pytest.raises(types.PackageError):
            _ = package.get_relative_path()
    else:
        assert package.get_relative_path() == result



import os
import uuid
from datetime import datetime

import metsrw
import pytest

from AIPscan import db
from AIPscan import test_helpers
from AIPscan.Aggregator import database_helpers
from AIPscan.Aggregator import types
from AIPscan.conftest import ORIGIN_PIPELINE
from AIPscan.conftest import STORAGE_LOCATION_1_CURRENT_LOCATION
from AIPscan.models import AIP
from AIPscan.models import Agent
from AIPscan.models import FetchJob
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import Pipeline
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService

FIXTURES_DIR = "fixtures"

TEST_SHA_256 = "79c16fa9573ec46c5f60fd54b34f314159e0623ca53d8d2f00c5875dbb4e0dfd"

BASE_FILE_DICT = {
    "name": "newfile.ext",
    "filepath": "/path/to/newfile.ext",
    "uuid": str(uuid.uuid4()),
    "date_created": datetime.now(),
    "puid": "fmt/1",
    "file_format": "Test Format",
    "size": "12345",
    "format_version": "1",
    "checksum_type": "sha-256",
    "checksum_value": "123456",
}

ORIGINAL_FILE_DICT = BASE_FILE_DICT.copy()
ORIGINAL_FILE_DICT["file_type"] = FileType.original

PRESERVATION_FILE_DICT = BASE_FILE_DICT.copy()
PRESERVATION_FILE_DICT["file_type"] = FileType.preservation
PRESERVATION_FILE_DICT["related_uuid"] = str(uuid.uuid4())


def test_create_storage_location_object(app_instance):
    LOCATION_UUID = str(uuid.uuid4())
    CURRENT_LOCATION = f"/api/v2/location/{LOCATION_UUID}"
    DESCRIPTION = "My test AIP Store location"
    storage_service = test_helpers.create_test_storage_service()

    database_helpers.create_storage_location_object(
        current_location=CURRENT_LOCATION,
        description=DESCRIPTION,
        storage_service_id=storage_service.id,
    )

    storage_location = StorageLocation.query.filter_by(
        current_location=CURRENT_LOCATION
    ).first()
    assert storage_location is not None
    assert storage_location.current_location == CURRENT_LOCATION
    assert storage_location.uuid == LOCATION_UUID
    assert storage_location.description == DESCRIPTION
    assert storage_location.storage_service_id == storage_service.id


def test_create_pipeline_object(app_instance):
    LOCATION_UUID = str(uuid.uuid4())
    ORIGIN_PIPELINE = f"/api/v2/pipeline/{LOCATION_UUID}"
    DASHBOARD_URL = "http://tessas-am-dashboard.example.com"

    database_helpers.create_pipeline_object(
        origin_pipeline=ORIGIN_PIPELINE, dashboard_url=DASHBOARD_URL
    )

    pipeline = Pipeline.query.filter_by(origin_pipeline=ORIGIN_PIPELINE).first()
    assert pipeline is not None
    assert pipeline.origin_pipeline == ORIGIN_PIPELINE
    assert pipeline.dashboard_url == DASHBOARD_URL


def test_create_aip(app_instance):
    """Test AIP creation."""
    PACKAGE_UUID = str(uuid.uuid4())
    TRANSFER_NAME = "some name"
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )
    pipeline = test_helpers.create_test_pipeline()

    database_helpers.create_aip_object(
        package_uuid=PACKAGE_UUID,
        transfer_name=TRANSFER_NAME,
        create_date="2020-11-02",
        mets_sha256=TEST_SHA_256,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
        origin_pipeline_id=pipeline.id,
    )

    aip = AIP.query.filter_by(uuid=PACKAGE_UUID).first()
    assert aip is not None
    assert aip.transfer_name == TRANSFER_NAME
    assert aip.storage_service_id == storage_service.id
    assert aip.fetch_job_id == fetch_job.id
    assert aip.origin_pipeline_id == pipeline.id


def test_delete_aip(app_instance):
    """Test AIP deletion."""
    PACKAGE_UUID = str(uuid.uuid4())
    storage_service = test_helpers.create_test_storage_service()
    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id
    )
    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )
    pipeline = test_helpers.create_test_pipeline()

    database_helpers.create_aip_object(
        package_uuid=PACKAGE_UUID,
        transfer_name="some name",
        create_date="2020-11-02",
        mets_sha256=TEST_SHA_256,
        storage_service_id=storage_service.id,
        storage_location_id=storage_location.id,
        fetch_job_id=fetch_job.id,
        origin_pipeline_id=pipeline.id,
    )

    aip = AIP.query.filter_by(uuid=PACKAGE_UUID).first()
    assert aip is not None

    database_helpers.delete_aip_object(aip)
    aip = AIP.query.filter_by(uuid=PACKAGE_UUID).first()
    assert aip is None


@pytest.mark.parametrize(
    "fixture_path, event_count, agent_link_multiplier",
    [
        (os.path.join("features_mets", "features-mets.xml"), 0, 0),
        (os.path.join("iso_mets", "iso_mets.xml"), 17, 3),
        (os.path.join("images_mets", "images.xml"), 76, 3),
    ],
)
def test_event_creation(
    app_instance, mocker, fixture_path, event_count, agent_link_multiplier
):
    """Make sure that we're seeing all of the events associated with
    an AIP and that they are potentially written to the database okay.
    Make sure too that the event_agent_relationship is established.
    """
    script_dir = os.path.dirname(os.path.realpath(__file__))
    mets_file = os.path.join(script_dir, FIXTURES_DIR, fixture_path)
    mets = metsrw.METSDocument.fromfile(mets_file)
    mocker.patch("AIPscan.models.Event")
    agent_find_match = mocker.patch(
        "AIPscan.Aggregator.database_helpers._create_agent_type_id"
    )
    mocker.patch(
        "sqlalchemy.orm.query.Query.first",
        return_value=Agent(
            linking_type_value="some_type_value",
            agent_type="an_agent_type",
            agent_value="an_agent_value",
            storage_service_id="1",
        ),
    )
    mocked_events = mocker.patch("AIPscan.db.session.add")
    mocker.patch("AIPscan.db.session.commit")
    for fsentry in mets.all_files():
        database_helpers.create_event_objects(fsentry, "some_id")
    assert mocked_events.call_count == event_count
    assert agent_find_match.call_count == event_count * agent_link_multiplier


@pytest.mark.parametrize(
    "fixture_path, number_of_unique_agents",
    [
        (os.path.join("features_mets", "features-mets.xml"), 0),
        (os.path.join("features_mets", "features-mets-added-agents.xml"), 5),
        (os.path.join("iso_mets", "iso_mets.xml"), 3),
        (os.path.join("images_mets", "images.xml"), 3),
    ],
)
def test_collect_agents(app_instance, fixture_path, number_of_unique_agents):
    """Make sure that we retrieve only unique Agents from the METS to
    then add to the database. Agents are "repeated" per PREMIS:OBJECT
    in METS.
    """
    script_dir = os.path.dirname(os.path.realpath(__file__))
    mets_file = os.path.join(script_dir, FIXTURES_DIR, fixture_path)
    mets = metsrw.METSDocument.fromfile(mets_file)
    agents = database_helpers.collect_mets_agents(mets)
    assert len(agents) == number_of_unique_agents


@pytest.mark.parametrize(
    "file_type, file_dict, is_original, connected_to_original",
    [
        # Test original file.
        (FileType.original, ORIGINAL_FILE_DICT, True, False),
        # Test preservation file tied to original.
        (FileType.preservation, PRESERVATION_FILE_DICT, False, True),
        # Test preservation file not tied to original.
        (FileType.preservation, PRESERVATION_FILE_DICT, False, False),
    ],
)
def test_create_file_object(
    app_with_populated_files,
    mocker,
    file_type,
    file_dict,
    is_original,
    connected_to_original,
):
    """Test adding files to database."""
    aip = AIP.query.first()
    first_original_file = File.query.filter_by(
        aip_id=aip.id, file_type=FileType.original.value
    ).first()

    get_file_props = mocker.patch(
        "AIPscan.Aggregator.database_helpers._get_file_properties"
    )
    get_file_props.return_value = file_dict

    get_original_file = mocker.patch(
        "AIPscan.Aggregator.database_helpers._get_original_file"
    )
    if not is_original:
        get_original_file.return_value = first_original_file

    mocker.patch("AIPscan.Aggregator.database_helpers.create_event_objects")
    add_normalization_date = mocker.patch(
        "AIPscan.Aggregator.database_helpers._add_normalization_date"
    )

    files_in_db = File.query.filter_by(aip_id=aip.id).all()
    assert len(files_in_db) == 2
    original_files = File.query.filter_by(
        aip_id=aip.id, file_type=FileType.original
    ).all()
    assert len(original_files) == 1
    preservation_files = File.query.filter_by(
        aip_id=aip.id, file_type=FileType.preservation
    ).all()
    assert len(preservation_files) == 1

    database_helpers.create_file_object(file_type, None, aip.id)

    files_in_db = File.query.filter_by(aip_id=aip.id).all()
    assert len(files_in_db) == 3
    if is_original:
        original_files = File.query.filter_by(
            aip_id=aip.id, file_type=FileType.original
        ).all()
        assert len(original_files) == 2
        preservation_files = File.query.filter_by(
            aip_id=aip.id, file_type=FileType.preservation
        ).all()
        assert len(preservation_files) == 1
        add_normalization_date.assert_not_called()
    else:
        original_files = File.query.filter_by(
            aip_id=aip.id, file_type=FileType.original
        ).all()
        assert len(original_files) == 1
        preservation_files = File.query.filter_by(
            aip_id=aip.id, file_type=FileType.preservation
        ).all()
        assert len(preservation_files) == 2
        add_normalization_date.assert_called_once()


@pytest.mark.parametrize(
    "current_location, storage_service_id, new_description, location_created",
    [
        # Create Storage Location if matching one doesn't exist.
        ("/api/v2/location/new-location/", 1, "", True),
        # Update Storage Location description if already exists.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, 1, "updated description", False),
    ],
)
def test_create_or_update_storage_location(
    storage_locations,
    mocker,
    current_location,
    storage_service_id,
    new_description,
    location_created,
):
    """Test that Storage Locations are created or updated as expected."""
    make_request = mocker.patch("AIPscan.Aggregator.tasks.make_request")
    make_request.return_value = {"description": new_description}

    create_storage_location_object = mocker.patch(
        "AIPscan.Aggregator.database_helpers.create_storage_location_object"
    )

    storage_service = db.session.get(StorageService, 1)

    storage_location = database_helpers.create_or_update_storage_location(
        current_location=current_location, storage_service=storage_service
    )

    if location_created:
        create_storage_location_object.assert_called_once()
    else:
        assert storage_location.description == new_description


@pytest.mark.parametrize(
    "origin_pipeline, new_url, pipeline_created",
    [
        # Create Pipeline if matching one doesn't exist.
        ("/api/v2/pipeline/new-pipeline/", "", True),
        # Update Pipeline's dashboard URL if already exists.
        (ORIGIN_PIPELINE, "http://new-url.example.com", False),
    ],
)
def test_create_or_update_pipeline(
    storage_locations, mocker, origin_pipeline, new_url, pipeline_created
):
    """Test that Storage Locations are created or updated as expected."""
    make_request = mocker.patch("AIPscan.Aggregator.tasks.make_request")
    make_request.return_value = {"remote_name": new_url}

    create_pipeline_object = mocker.patch(
        "AIPscan.Aggregator.database_helpers.create_pipeline_object"
    )

    get_storage_service_api_url = mocker.patch(
        "AIPscan.Aggregator.database_helpers.get_storage_service_api_url"
    )
    get_storage_service_api_url.return_value = (None, None)

    storage_service = db.session.get(StorageService, 1)

    pipeline = database_helpers.create_or_update_pipeline(
        origin_pipeline=origin_pipeline, storage_service=storage_service
    )

    if pipeline_created:
        create_pipeline_object.assert_called_once()
    else:
        assert pipeline.dashboard_url == new_url


def test_create_fetch_job(app_instance, mocker, tmp_path):
    datetime_obj_start = datetime.now().replace(microsecond=0)
    timestamp_str = datetime_obj_start.strftime("%Y-%m-%d-%H-%M-%S")
    storage_service_id = 1

    mocker.patch("AIPscan.db.session.add")
    mocker.patch("AIPscan.db.session.commit")

    download_root = tmp_path / "aipscan-downloads"
    mocker.patch(
        "AIPscan.Aggregator.database_helpers.get_download_root",
        return_value=os.fspath(download_root),
    )

    fetch_job = database_helpers.create_fetch_job(
        datetime_obj_start, timestamp_str, storage_service_id
    )
    expected_directory = os.path.join(os.fspath(download_root), timestamp_str, "")
    assert fetch_job.total_packages is None
    assert fetch_job.total_deleted_aips is None
    assert fetch_job.total_aips is None
    assert fetch_job.download_start == datetime_obj_start
    assert fetch_job.download_end is None
    assert fetch_job.download_directory == expected_directory
    assert fetch_job.storage_service_id == storage_service_id


def test_update_fetch_job(app_instance, mocker):
    fetch_job = FetchJob(
        total_packages=0,
        total_aips=0,
        total_deleted_aips=0,
        download_start=datetime.now(),
        download_end=datetime.now(),
        download_directory="/some/directory/",
        storage_service_id=1,
    )

    mocker.patch("sqlalchemy.orm.query.Query.first", return_value=fetch_job)
    mocker.patch("AIPscan.db.session.commit")

    # Define processed packages
    deleted = types.StorageServicePackage()
    deleted.deleted = True

    replica = types.StorageServicePackage()
    replica.replica = True
    replica.aip = True

    aip = types.StorageServicePackage()
    aip.aip = True

    dip = types.StorageServicePackage()
    dip.dip = True

    sip = types.StorageServicePackage()
    sip.sip = True

    processed_packages = [deleted, replica, aip, dip, sip]

    # Only some packags may have been processed
    total_packages_count = 10

    obj = database_helpers.update_fetch_job(1, processed_packages, total_packages_count)

    assert obj.total_packages == total_packages_count
    assert obj.total_deleted_aips == 1
    assert obj.total_replicas == 1
    assert obj.total_aips == 1
    assert obj.total_dips == 1
    assert obj.total_sips == 1



import hashlib
import logging
import os
import time
import uuid
from datetime import datetime
from io import StringIO

import tzlocal

from AIPscan import db
from AIPscan.models import AIP
from AIPscan.models import Agent
from AIPscan.models import Event
from AIPscan.models import EventAgent
from AIPscan.models import FetchJob
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import Pipeline
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService
from AIPscan.models import index_tasks

TEST_SHA_256 = "79c16fa9573ec46c5f60fd54b34f314159e0623ca53d8d2f00c5875dbb4e0dfd"


def file_sha256_hash(filepath):
    """Return SHA256 hash for contents of file."""
    with open(filepath, "rb") as f:
        return hashlib.sha256(f.read()).hexdigest()


def _add_test_object_to_db(test_object):
    """Add and commit test object to database."""
    db.session.add(test_object)
    db.session.commit()


def create_test_storage_service(**kwargs):
    """Create and return a Storage Service with overridable defaults."""
    name = kwargs.get("name", "test storage service")
    existing = StorageService.query.filter_by(name=name).first()
    if existing is not None:
        return existing
    if "url" in kwargs:
        url = kwargs["url"]
    elif "name" in kwargs:
        url = kwargs["name"]
    else:
        url = "http://example.com"
    storage_service = StorageService(
        name=name,
        url=url,
        user_name=kwargs.get("user_name", "test"),
        api_key=kwargs.get("api_key", "test"),
        download_limit=kwargs.get("download_limit", 0),
        download_offset=kwargs.get("download_offset", 10),
        default=kwargs.get("default", True),
    )
    _add_test_object_to_db(storage_service)
    return storage_service


def create_test_fetch_job(**kwargs):
    """Create and return a test Fetch Job with overridable defaults."""
    storage_service_id = kwargs.get("storage_service_id")
    if not storage_service_id:
        storage_service = create_test_storage_service()
        storage_service_id = storage_service.id
    fetch_job = FetchJob(
        total_packages=kwargs.get("total_packages", 1),
        total_aips=kwargs.get("total_aips", 1),
        total_deleted_aips=kwargs.get("total_deleted_aips", 0),
        download_start=kwargs.get("download_start", datetime.now()),
        download_end=kwargs.get("download_end", datetime.now()),
        download_directory=kwargs.get("download_directory", "/some/directory/"),
        storage_service_id=storage_service_id,
    )
    _add_test_object_to_db(fetch_job)
    return fetch_job


def create_test_storage_location(**kwargs):
    """Create and return a test Storage Location with overridable defaults."""
    storage_service_id = kwargs.get("storage_service_id")
    if storage_service_id is None:
        storage_service_id = kwargs.get("default_storage_service_id")
    if not storage_service_id:
        storage_service = create_test_storage_service()
        storage_service_id = storage_service.id
    current_location = kwargs.get(
        "current_location", f"/api/v2/location/{uuid.uuid4()}/"
    )
    storage_location = StorageLocation(
        current_location=current_location,
        description=kwargs.get("description", "test storage location"),
        storage_service_id=storage_service_id,
    )
    _add_test_object_to_db(storage_location)
    return storage_location


def create_test_pipeline(**kwargs):
    """Create and return a test Pipeline with overridable defaults."""
    origin_pipeline = kwargs.get("origin_pipeline", f"/api/v2/pipeline/{uuid.uuid4()}")
    pipeline = Pipeline(
        origin_pipeline=origin_pipeline,
        dashboard_url=kwargs.get("dashboard_url", "http://example.com"),
    )
    _add_test_object_to_db(pipeline)
    return pipeline


def create_test_aip(**kwargs):
    """Create and return a test AIP with overridable defaults."""
    storage_service_id = kwargs.get("storage_service_id")
    if not storage_service_id:
        storage_service = create_test_storage_service()
        storage_service_id = storage_service.id
    storage_location_id = kwargs.get("storage_location_id")
    if not storage_location_id:
        storage_location = create_test_storage_location(
            storage_service_id=storage_service_id
        )
        storage_location_id = storage_location.id
    fetch_job_id = kwargs.get("fetch_job_id")
    if not fetch_job_id:
        fetch_job = create_test_fetch_job(storage_service_id=storage_service_id)
        fetch_job_id = fetch_job.id
    origin_pipeline_id = kwargs.get("origin_pipeline_id")
    if not origin_pipeline_id:
        pipeline = create_test_pipeline()
        origin_pipeline_id = pipeline.id

    aip = AIP(
        uuid=kwargs.get("uuid", str(uuid.uuid4())),
        transfer_name=kwargs.get("transfer_name", "Test AIP"),
        create_date=kwargs.get("create_date", datetime.now()),
        mets_sha256=kwargs.get("mets_sha256", TEST_SHA_256),
        size=kwargs.get("size", 100),
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job_id,
        origin_pipeline_id=origin_pipeline_id,
    )
    _add_test_object_to_db(aip)
    return aip


def create_test_file(**kwargs):
    """Create and return a test File with overridable defaults."""
    aip_id = kwargs.get("aip_id")
    if not aip_id:
        existing_aip = AIP.query.first()
        if existing_aip is not None:
            aip_id = existing_aip.id
        else:
            aip = create_test_aip()
            aip_id = aip.id
    file_ = File(
        name=kwargs.get("name", "file_name.ext"),
        filepath=kwargs.get("filepath", "/path/to/file_name.ext"),
        uuid=kwargs.get("uuid", "222222222222-2222-2222-22222222"),
        file_type=kwargs.get("file_type", FileType.original),
        size=kwargs.get("size", 0),
        date_created=kwargs.get("date_created", datetime.now()),
        puid=kwargs.get("puid", "fmt/test-1"),
        file_format=kwargs.get("file_format", "ACME File Format"),
        format_version=kwargs.get("format_version", "0.0.0"),
        checksum_type=kwargs.get("checksum_type", "test"),
        checksum_value=kwargs.get("checksum_value", "test"),
        original_file_id=kwargs.get("original_file_id", None),
        aip_id=aip_id,
    )
    _add_test_object_to_db(file_)
    return file_


def create_test_agent(**kwargs):
    """Create and return a test Agent with overridable defaults."""
    storage_service_id = kwargs.get("storage_service_id")
    if not storage_service_id:
        # Prefer an existing storage service if present so tests that
        # set one up in advance get agents linked to it.
        existing = StorageService.query.order_by(StorageService.id.asc()).first()
        if existing is None:
            existing = create_test_storage_service()
        storage_service_id = existing.id
    agent = Agent(
        linking_type_value=kwargs.get("linking_type_value", "Archivematica user pk-1"),
        agent_type=kwargs.get("agent_type", "Archivematica user"),
        agent_value=kwargs.get(
            "agent_value", 'username="user one", first_name="", last_name=""'
        ),
        storage_service_id=storage_service_id,
    )
    _add_test_object_to_db(agent)
    return agent


def create_test_event(**kwargs):
    """Create and return a test Event with overridable defaults."""
    event = Event(
        event_type=kwargs.get("event_type", kwargs.get("type", "ingestion")),
        uuid=kwargs.get("uuid", str(uuid.uuid4())),
        date=kwargs.get("date", datetime.now()),
        detail=kwargs.get("detail", "ingestion detail"),
        outcome=kwargs.get("outcome", "success"),
        outcome_detail=kwargs.get("outcome_detail", "outcome detail"),
        file_id=kwargs.get("file_id", 1),
    )
    _add_test_object_to_db(event)
    return event


def create_test_event_agent(**kwargs):
    """Create and return a test EventAgent relationship with overridable defaults."""
    event_relationship = EventAgent.insert().values(
        event_id=kwargs.get("event_id", 1), agent_id=kwargs.get("agent_id", 1)
    )
    db.session.execute(event_relationship)
    db.session.commit()
    return event_relationship


def create_test_index_tasks(fetch_job_id, task_id):
    index_task_obj = index_tasks(
        index_task_id=task_id, fetch_job_id=fetch_job_id, indexing_start=datetime.now()
    )
    db.session.add(index_task_obj)
    db.session.commit()


def set_timezone(new_tz):
    os.environ["TZ"] = new_tz
    time.tzset()


def set_timezone_and_return_current_timezone(new_tz):
    local_timezone = tzlocal.get_localzone_name()
    set_timezone(new_tz)

    return local_timezone


def add_logger_streamer(logger):
    """Add stream handler to logger and return it."""
    logger.setLevel(logging.DEBUG)

    log_streamer = StringIO()
    handler = logging.StreamHandler(log_streamer)
    logger.addHandler(handler)

    return log_streamer



class PackageError(Exception):
    """There are things we cannot do with the package type unless it
    is completed properly. Let the user know.
    """


class StorageServicePackage:
    """Type that can record information about a storage service package
    and provide helpers as to whether or not we should process it.
    """

    default_pair_tree = "0000-0000-0000-0000-0000-0000-0000-0000-"
    compressed_ext = ".7z"

    def __init__(self, **kwargs):
        """Package constructor"""

        DELETED = "deleted"
        REPLICA = "replica"
        AIP = "aip"
        DIP = "dip"
        SIP = "sip"

        UUID = "uuid"
        CURRENT_LOCATION = "current_location"
        CURRENT_PATH = "current_path"
        ORIGIN_PIPELINE = "origin_pipeline"

        self.deleted = False
        self.replica = False
        self.aip = False
        self.dip = False
        self.sip = False
        self.uuid = None
        self.current_location = None
        self.current_path = None
        self.origin_pipeline = None

        if kwargs:
            self.deleted = kwargs.get(DELETED, self.deleted)
            self.replica = kwargs.get(REPLICA, self.replica)
            self.aip = kwargs.get(AIP, self.aip)
            self.dip = kwargs.get(DIP, self.dip)
            self.sip = kwargs.get(SIP, self.sip)
            self.uuid = kwargs.get(UUID, self.uuid)
            self.current_location = kwargs.get(CURRENT_LOCATION, self.current_location)
            self.current_path = kwargs.get(CURRENT_PATH, self.uuid)
            self.origin_pipeline = kwargs.get(ORIGIN_PIPELINE, self.origin_pipeline)

    def __repr__(self):
        ret = f"aip: '{self.aip}'; dip: '{self.dip}'; sip: '{self.sip}'; deleted: '{self.deleted}'; replica: '{self.replica}';"
        return ret

    def __eq__(self, other):
        """Comparison operator"""
        ret = True
        if self.aip != other.aip:
            ret = False
        if self.dip != other.dip:
            ret = False
        if self.sip != other.sip:
            ret = False
        if self.deleted != other.deleted:
            ret = False
        if self.replica != other.replica:
            ret = False
        if self.uuid != other.uuid:
            ret = False
        if self.current_location != other.current_location:
            ret = False
        if self.current_path != other.current_path:
            ret = False
        if self.origin_pipeline != other.origin_pipeline:
            ret = False
        return ret

    def is_aip(self):
        """Determine whether the package is a AIP"""
        if (
            self.aip
            and not self.deleted
            and not self.replica
            and not self.dip
            and not self.sip
        ):
            return True
        return False

    def is_undeleted_aip(self):
        if self.is_deleted():
            return False

        return self.is_aip()

    def is_dip(self):
        """Determine whether the package is a DIP"""
        if (
            self.dip
            and not self.deleted
            and not self.replica
            and not self.aip
            and not self.sip
        ):
            return True
        return False

    def is_sip(self):
        """Determine whether the package is a SIP"""
        if (
            self.sip
            and not self.deleted
            and not self.replica
            and not self.dip
            and not self.aip
        ):
            return True
        return False

    def is_replica(self):
        """Determine whether the package is a replica package"""
        if (
            self.replica
            and self.aip
            and not self.deleted
            and not self.sip
            and not self.dip
        ):
            return True
        return False

    def is_deleted(self):
        """Determine whether the package is a deleted package"""
        if self.deleted:
            return True
        return False

    def get_relative_path(self):
        """Return relative path from current_path."""
        if self.dip or self.sip or self.replica:
            raise PackageError(
                "Get relative path for sip or replica packages not yet implemented"
            )
        if self.deleted:
            raise PackageError("There are no relative paths for deleted packages")
        if self.uuid is None:
            raise PackageError("Cannot generate a relative path without a package UUID")
        rel = ""
        left_offset = len(self.default_pair_tree)
        right_offset = -len(self.compressed_ext)
        try:
            if self.current_path.endswith(self.compressed_ext):
                rel = self.current_path[left_offset:right_offset]
            else:
                rel = self.current_path[left_offset:]
        except AttributeError as err:
            raise PackageError("Current path doesn't exist for the package") from err
        return f"{rel}/data/METS.{self.uuid}.xml"



from flask_wtf import FlaskForm
from wtforms import BooleanField
from wtforms import PasswordField
from wtforms import StringField
from wtforms import SubmitField
from wtforms.validators import DataRequired


class LoginForm(FlaskForm):
    username = StringField("Username", validators=[DataRequired()])
    password = PasswordField("Password", validators=[DataRequired()])
    remember_me = BooleanField("Remember Me")
    submit = SubmitField("Sign In")



import json
import logging
import os
import uuid
from datetime import datetime

import celery
import pytest

from AIPscan import db
from AIPscan import test_helpers
from AIPscan.Aggregator.tasks import TaskError
from AIPscan.Aggregator.tasks import delete_aip
from AIPscan.Aggregator.tasks import delete_fetch_job
from AIPscan.Aggregator.tasks import delete_storage_service
from AIPscan.Aggregator.tasks import get_mets
from AIPscan.Aggregator.tasks import handle_deletion
from AIPscan.Aggregator.tasks import index_task
from AIPscan.Aggregator.tasks import make_request
from AIPscan.Aggregator.tasks import parse_package_list_file
from AIPscan.Aggregator.tasks import process_packages
from AIPscan.Aggregator.tasks import start_index_task
from AIPscan.Aggregator.tests import INVALID_JSON
from AIPscan.Aggregator.tests import REQUEST_URL
from AIPscan.Aggregator.tests import REQUEST_URL_WITHOUT_API_KEY
from AIPscan.Aggregator.tests import RESPONSE_DICT
from AIPscan.Aggregator.tests import VALID_JSON
from AIPscan.Aggregator.tests import MockResponse
from AIPscan.Aggregator.types import StorageServicePackage
from AIPscan.models import AIP
from AIPscan.models import Agent
from AIPscan.models import FetchJob
from AIPscan.models import StorageService
from AIPscan.models import index_tasks

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURES_DIR = os.path.join(SCRIPT_DIR, "fixtures")


def _get_aips(storage_service_id):
    """Return queryset of AIPs associated with Storage Service ID."""
    return AIP.query.filter_by(storage_service_id=storage_service_id).all()


@pytest.mark.parametrize(
    "fixture_path, package_uuid",
    [
        # First METS file.
        (
            os.path.join("features_mets", "features-mets.xml"),
            "ab793f82-27b0-4c2e-ac3e-621bab5af8f1",
        ),
        # Another METS file.
        (
            os.path.join("images_mets", "images.xml"),
            "2d718ecf-828a-4487-83ce-873cf9edca1a",
        ),
    ],
)
def test_get_mets_task(app_instance, tmpdir, mocker, fixture_path, package_uuid):
    """Test that fetch jobs after the first don't duplicate content."""
    mets_file = os.path.join(FIXTURES_DIR, fixture_path)

    def mock_download_mets(
        storage_service,
        package_uuid,
        relative_path_to_mets,
        timestamp_str,
        package_list_no,
    ):
        return mets_file, test_helpers.file_sha256_hash(mets_file)

    mocker.patch("AIPscan.Aggregator.tasks.download_mets", mock_download_mets)
    delete_mets_file = mocker.patch("AIPscan.Aggregator.tasks.os.remove")

    storage_service = test_helpers.create_test_storage_service()
    storage_service_id = storage_service.id

    storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service_id
    )
    storage_location_id = storage_location.id

    pipeline = test_helpers.create_test_pipeline(storage_service_id=storage_service_id)
    pipeline_id = pipeline.id

    get_storage_location = mocker.patch(
        "AIPscan.Aggregator.database_helpers.create_or_update_storage_location"
    )
    get_storage_location.return_value = storage_location

    # No AIPs should exist at this point.
    aips = _get_aips(storage_service_id)
    assert not aips

    # Set up custom logger and add handler to capture output
    customlogger = logging.getLogger(__name__)
    log_stream = test_helpers.add_logger_streamer(customlogger)

    # Create AIP and verify record.
    fetch_job1 = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service_id
    )
    fetch_job1_id = fetch_job1.id
    get_mets(
        package_uuid=package_uuid,
        aip_size=1000,
        relative_path_to_mets="test",
        timestamp_str=datetime.now()
        .replace(microsecond=0)
        .strftime("%Y-%m-%d-%H-%M-%S"),
        package_list_no=1,
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job1_id,
        origin_pipeline_id=pipeline_id,
        customlogger=customlogger,
    )
    aips = _get_aips(storage_service_id)
    assert len(aips) == 1
    fetch_job1_refreshed = db.session.get(FetchJob, fetch_job1_id)
    assert len(fetch_job1_refreshed.aips) == 1

    original_mets_sha256 = aips[0].mets_sha256

    # Try to create AIP again and verify no duplicate is created.
    fetch_job2 = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service_id
    )
    fetch_job2_id = fetch_job2.id
    get_mets(
        package_uuid=package_uuid,
        aip_size=1000,
        relative_path_to_mets="test",
        timestamp_str=datetime.now()
        .replace(microsecond=0)
        .strftime("%Y-%m-%d-%H-%M-%S"),
        package_list_no=1,
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job2_id,
        origin_pipeline_id=pipeline_id,
    )
    aips = _get_aips(storage_service_id)
    assert len(aips) == 1
    assert aips[0].mets_sha256 == original_mets_sha256
    fetch_job1_refreshed = db.session.get(FetchJob, fetch_job1_id)
    fetch_job2_refreshed = db.session.get(FetchJob, fetch_job2_id)
    assert len(fetch_job1_refreshed.aips) == 1
    assert len(fetch_job2_refreshed.aips) == 0

    # Replace METS with a new METS file and run again. The old AIP record
    # should be deleted and replaced with one from the new METS.
    mets_file = os.path.join(FIXTURES_DIR, "iso_mets", "iso_mets.xml")
    fetch_job3 = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service_id
    )
    fetch_job3_id = fetch_job3.id
    get_mets(
        package_uuid=package_uuid,
        aip_size=1000,
        relative_path_to_mets="test",
        timestamp_str=datetime.now()
        .replace(microsecond=0)
        .strftime("%Y-%m-%d-%H-%M-%S"),
        package_list_no=1,
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job3_id,
        origin_pipeline_id=pipeline_id,
    )
    aips = _get_aips(storage_service_id)
    assert len(aips) == 1
    assert aips[0].mets_sha256 != original_mets_sha256
    fetch_job1_refreshed = db.session.get(FetchJob, fetch_job1_id)
    fetch_job2_refreshed = db.session.get(FetchJob, fetch_job2_id)
    fetch_job3_refreshed = db.session.get(FetchJob, fetch_job3_id)
    assert len(fetch_job1_refreshed.aips) == 0
    assert len(fetch_job2_refreshed.aips) == 0
    assert len(fetch_job3_refreshed.aips) == 1

    delete_calls = [
        mocker.call(os.path.join(FIXTURES_DIR, fixture_path)),
        mocker.call(mets_file),
    ]
    delete_mets_file.assert_has_calls(delete_calls, any_order=True)
    assert delete_mets_file.call_count == 3

    # Test that custom logger was used
    assert (
        log_stream.getvalue()
        == f"Processing METS file {os.path.basename(fixture_path)}\n"
    )


def test_delete_fetch_job_task(app_instance, tmpdir, mocker):
    """Test that fetch job gets deleted by delete fetch job task logic."""
    storage_service = test_helpers.create_test_storage_service()

    # Create fetch job and confirm expected ID
    fetch_job1 = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )
    assert fetch_job1.id == 1

    # Create index_tasks instance and confirm expected ID
    task_id = celery.uuid()
    test_helpers.create_test_index_tasks(fetch_job1.id, task_id)

    index_tasks_obj = index_tasks.query.filter_by(fetch_job_id=fetch_job1.id).first()
    assert index_tasks_obj.index_task_id == task_id

    # Delete fetch job and confirm it no longer exists
    delete_fetch_job(fetch_job1.id)

    assert FetchJob.query.filter_by(id=fetch_job1.id).first() is None

    index_tasks_obj = index_tasks.query.filter_by(fetch_job_id=fetch_job1.id).first()
    assert index_tasks_obj is None


def test_delete_storage_service_task(app_instance, tmpdir, mocker):
    """Test that storage service gets deleted by delete storage service job task logic."""
    storage_service = test_helpers.create_test_storage_service()
    test_helpers.create_test_agent()

    assert Agent.query.filter_by(storage_service_id=storage_service.id).count() == 1

    deleted_ss = StorageService.query.filter_by(id=storage_service.id).first()
    assert deleted_ss is not None

    delete_storage_service(storage_service.id)

    deleted_ss = StorageService.query.filter_by(id=storage_service.id).first()
    assert deleted_ss is None
    assert Agent.query.filter_by(storage_service_id=storage_service.id).count() == 0


@pytest.mark.parametrize(
    "response, raises_task_error",
    [
        # Test 200 response with valid JSON.
        (MockResponse(200, VALID_JSON), False),
        # Test 500 response raises TaskError.
        (MockResponse(500, VALID_JSON), True),
        # Test 200 response with invalid JSON raises TaskError.
        (MockResponse(200, INVALID_JSON), True),
    ],
)
def test_make_request(mocker, response, raises_task_error):
    """Test handling of Storage Service response."""
    request = mocker.patch("AIPscan.Aggregator.tasks.requests.get")
    request.return_value = response

    if raises_task_error:
        with pytest.raises(TaskError):
            _ = make_request(REQUEST_URL, REQUEST_URL_WITHOUT_API_KEY)
    else:
        return_dict = make_request(REQUEST_URL, REQUEST_URL_WITHOUT_API_KEY)
        assert return_dict["key"] == RESPONSE_DICT["key"]


def test_parse_package_list_file(tmpdir):
    """Test that JSON package list files are being parsed."""
    json_file_path = tmpdir.join("packages.json")
    json_file_path.write(json.dumps({"objects": []}))

    package = parse_package_list_file(json_file_path, None, True)
    assert "objects" in package

    package_list = package["objects"]
    assert len(package_list) == 0


def test_process_packages_json_file_deletion(app_instance, tmpdir, mocker):
    """Test that JSON package lists are deleted after being parsed."""
    json_file_path = tmpdir.join("packages.json")
    json_file_path.write(json.dumps({"objects": []}))

    delete_package_json = mocker.patch("AIPscan.Aggregator.tasks.os.remove")

    package_list = parse_package_list_file(json_file_path, None, True)

    process_packages(package_list, 1, str(datetime.now()), 1, 1, True)

    delete_package_json.assert_called_with(json_file_path)


def test_process_packages(app_instance, tmpdir, mocker):
    """Test that JSON package lists are deleted after being parsed."""
    aip_package_uuid = str(uuid.uuid4())
    aip_package_data = {
        "uuid": aip_package_uuid,
        "package_type": "AIP",
        "current_path": str(tmpdir),
    }

    deleted_sip_package_uuid = str(uuid.uuid4())
    deleted_sip_package_data = {
        "uuid": deleted_sip_package_uuid,
        "package_type": "SIP",
        "current_path": str(tmpdir),
        "deleted": True,
    }

    # Write test AIP to JSON file (from which to general a list of packages)
    json_file_path = tmpdir.join("packages.json")
    json_file_path.write(
        json.dumps({"objects": [aip_package_data, deleted_sip_package_data]})
    )

    # Get test package list
    package_list = parse_package_list_file(json_file_path, None, True)

    # Process test package list
    mocker.patch(
        "AIPscan.Aggregator.database_helpers.create_or_update_storage_location"
    )
    mocker.patch("AIPscan.Aggregator.database_helpers.create_or_update_pipeline")

    # Set up custom logger and add handler to capture output
    customlogger = logging.getLogger(__name__)
    log_stream = test_helpers.add_logger_streamer(customlogger)

    processed_packages = process_packages(
        package_list, 1, str(datetime.now()), 1, 1, False, customlogger, 1, 1
    )

    # Test that custom logger was used
    assert log_stream.getvalue() == f"Processing {aip_package_uuid} (1 of 1)\n"

    # Make sure only one package was processed and that is was the non-deleted AIP
    assert len(processed_packages) == 1
    assert processed_packages[0].aip is True


def test_handle_deletion(app_instance, mocker):
    """Test that delete handler handles deletion correctly."""
    PACKAGE_UUID = str(uuid.uuid4())

    # Make sure package deleted on storage service gets deleted locally
    package = StorageServicePackage(uuid=PACKAGE_UUID, deleted=True)
    mock_delete_aip = mocker.patch("AIPscan.Aggregator.tasks.delete_aip")

    handle_deletion(package)

    mock_delete_aip.assert_called_with(PACKAGE_UUID)

    # Make sure package not deleted on storage service doesn't get deleted
    package = StorageServicePackage(uuid=PACKAGE_UUID, deleted=False)
    mock_delete_aip = mocker.patch("AIPscan.Aggregator.tasks.delete_aip")

    handle_deletion(package)

    mock_delete_aip.assert_not_called()


def test_delete_aip(app_instance):
    """Test that SS deleted AIPs get removed from the database."""
    PACKAGE_UUID = str(uuid.uuid4())

    test_helpers.create_test_aip(uuid=PACKAGE_UUID)

    # Make sure test AIP exists before deletion
    deleted_aip = AIP.query.filter_by(uuid=PACKAGE_UUID).first()
    assert deleted_aip is not None

    delete_aip(PACKAGE_UUID)

    # Make sure test AIP doesn't exist after deletion
    deleted_aip = AIP.query.filter_by(uuid=PACKAGE_UUID).first()
    assert deleted_aip is None


def test_start_index_task(app_instance, mocker):
    """Test that index_tasks record gets create by start index job function logic."""
    # Mock creation of index task
    mock_index_task = mocker.patch("AIPscan.Aggregator.tasks.index_task.delay")

    class MockTask:
        id = 1

    mock_index_task.return_value = MockTask()

    # Create test fetch job
    storage_service = test_helpers.create_test_storage_service()
    storage_service_id = storage_service.id

    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service_id
    )
    fetch_job_id = fetch_job.id

    # At this point no index_tasks should exist for the fetch job
    index_task_obj = index_tasks.query.filter_by(fetch_job_id=fetch_job_id).first()
    assert index_task_obj is None

    # Start mock index task, making sure it got started with the right value
    start_index_task(fetch_job_id)

    mock_index_task.assert_called_with(fetch_job_id)

    # At this point an index_tasks should exist for the fetch job
    index_task_obj = index_tasks.query.filter_by(fetch_job_id=fetch_job_id).first()
    assert type(index_task_obj) is index_tasks


def test_index_task(app_instance, enable_typesense, mocker):
    """Test index task."""
    # Create test fetch job
    storage_service = test_helpers.create_test_storage_service()

    fetch_job = test_helpers.create_test_fetch_job(
        storage_service_id=storage_service.id
    )

    # Mock call to index initialize function
    mocker.patch("AIPscan.typesense_helpers.initialize_index")

    # Create test index tasks object and test index task with fake Celery task ID
    task_id = celery.uuid()
    test_helpers.create_test_index_tasks(fetch_job.id, task_id)

    # Mock call to finish bulk document creation function
    mock_finish_bulk_doc = mocker.patch(
        "AIPscan.typesense_helpers.finish_bulk_document_creation"
    )

    # Start index task
    index_task.apply((fetch_job.id,), task_id=task_id)

    # Make sure finish bulk document creation function got called
    mock_finish_bulk_doc.assert_called()



"""Helpers for working with Aggregator download directories."""

import os
from importlib import resources

from flask import current_app

DEFAULT_DOWNLOAD_ROOT = os.fspath(resources.files(__package__).joinpath("downloads"))


def get_download_root(app=None) -> str:
    """Return the base directory used to stage download artifacts."""
    if app is not None:
        return app.config.get("AGGREGATOR_DOWNLOAD_ROOT", DEFAULT_DOWNLOAD_ROOT)

    try:
        application = current_app._get_current_object()
    except RuntimeError:
        env_override = os.getenv("AGGREGATOR_DOWNLOAD_ROOT")
        if env_override:
            return env_override
        return DEFAULT_DOWNLOAD_ROOT

    return application.config.get("AGGREGATOR_DOWNLOAD_ROOT", DEFAULT_DOWNLOAD_ROOT)



"""Application command-line interface helpers."""

import time

import click
import sqlalchemy as sa
from flask.cli import with_appcontext
from sqlalchemy.exc import SQLAlchemyError

from AIPscan import db
from AIPscan.models import StorageService

DEFAULT_DOWNLOAD_LIMIT = 20
DEFAULT_DOWNLOAD_OFFSET = 0
WRITE_RETRY_ATTEMPTS = 3
WRITE_RETRY_DELAY_SECONDS = 2


def register_cli(app):
    """Register custom CLI commands with the Flask app."""

    app.cli.add_command(storage_service_bootstrap)


@click.command("storage-service-bootstrap")
@click.option("--name", required=True, help="Storage Service name")
@click.option("--url", required=True, help="Storage Service base URL")
@click.option("--username", required=True, help="Storage Service API username")
@click.option("--api-key", required=True, help="Storage Service API key")
@click.option(
    "--default",
    "make_default",
    is_flag=True,
    default=False,
    help="Mark this Storage Service as the default",
)
@click.option(
    "--wait",
    is_flag=True,
    help="Retry database connection until it becomes available.",
)
@click.option(
    "--wait-timeout",
    type=int,
    default=60,
    show_default=True,
    help="Maximum seconds to wait for the database when --wait is set.",
)
@with_appcontext
def storage_service_bootstrap(
    name, url, username, api_key, make_default, wait, wait_timeout
):
    """Create or update a Storage Service from CLI parameters."""

    if wait:
        _wait_for_database(wait_timeout)

    attempts_remaining = WRITE_RETRY_ATTEMPTS
    last_exception = None

    while attempts_remaining > 0:
        try:
            storage_service = _load_storage_service(name)

            if storage_service:
                message = _update_storage_service(
                    storage_service, url, username, api_key, make_default
                )
                if message is None:
                    click.echo(f"Storage service '{name}' already up to date.")
                    return

                db.session.commit()
                click.echo(message)
                return

            message = _create_storage_service(
                name, url, username, api_key, make_default
            )
            db.session.commit()
            click.echo(message)
            return
        except SQLAlchemyError as exc:
            db.session.rollback()
            last_exception = exc
            attempts_remaining -= 1
            if attempts_remaining == 0:
                break
            _wait_before_retry(wait, wait_timeout)

    raise click.ClickException(
        "Failed to write storage service; database not ready?"
    ) from last_exception


def _clear_default_flags(exclude_ids=None):
    """Unset the default flag for all storage services except excluded ones."""

    exclude_ids = list(exclude_ids or [])

    stmt = sa.update(StorageService).values(default=False)
    if exclude_ids:
        stmt = stmt.where(StorageService.id.notin_(exclude_ids))

    db.session.execute(stmt)


def _load_storage_service(name):
    stmt = sa.select(StorageService).where(StorageService.name == name)
    return db.session.execute(stmt).scalar_one_or_none()


def _update_storage_service(storage_service, url, username, api_key, make_default):
    changes = []
    for attr, value in ("url", url), ("user_name", username), ("api_key", api_key):
        if getattr(storage_service, attr) != value:
            setattr(storage_service, attr, value)
            changes.append(attr)

    if make_default and not storage_service.default:
        _clear_default_flags(exclude_ids=[storage_service.id])
        storage_service.default = True
        changes.append("default")

    if not changes:
        return None

    return f"Storage service '{storage_service.name}' updated ({', '.join(changes)})."


def _create_storage_service(name, url, username, api_key, make_default):
    if make_default:
        _clear_default_flags()

    storage_service = StorageService(
        name=name,
        url=url,
        user_name=username,
        api_key=api_key,
        download_limit=DEFAULT_DOWNLOAD_LIMIT,
        download_offset=DEFAULT_DOWNLOAD_OFFSET,
        default=make_default,
    )
    db.session.add(storage_service)
    return f"Storage service '{name}' created."


def _wait_before_retry(wait_enabled, wait_timeout):
    delay = WRITE_RETRY_DELAY_SECONDS
    if wait_enabled:
        delay = min(wait_timeout, WRITE_RETRY_DELAY_SECONDS)

    time.sleep(max(1, delay))


def _wait_for_database(timeout_seconds):
    deadline = time.monotonic() + timeout_seconds
    while True:
        try:
            db.session.execute(sa.select(1))
            return
        except SQLAlchemyError as err:
            db.session.rollback()
            if time.monotonic() >= deadline:
                raise click.ClickException(
                    "Database not available within wait timeout."
                ) from err
            time.sleep(1)



"""Views contains the primary routes for navigation around AIPscan's
Reporter module. Reports themselves as siphoned off into separate module
files with singular responsibility for a report.
"""

from datetime import datetime

import requests
from flask import Response
from flask import abort
from flask import current_app
from flask import jsonify
from flask import make_response
from flask import render_template
from flask import request
from flask import session

from AIPscan import db
from AIPscan.Aggregator.task_helpers import get_mets_url
from AIPscan.models import AIP
from AIPscan.models import Event
from AIPscan.models import FetchJob
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.models import Pipeline
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService

# Flask's idiom requires code using routing decorators to be imported
# up-front. But that means it might not be called directly by a module.
from AIPscan.Reporter import report_aip_contents  # noqa: F401
from AIPscan.Reporter import report_aips_by_format  # noqa: F401
from AIPscan.Reporter import report_aips_by_puid  # noqa: F401
from AIPscan.Reporter import report_format_versions_count  # noqa: F401
from AIPscan.Reporter import report_formats_count  # noqa: F401
from AIPscan.Reporter import report_ingest_log  # noqa: F401
from AIPscan.Reporter import report_largest_aips  # noqa: F401
from AIPscan.Reporter import report_largest_files  # noqa: F401
from AIPscan.Reporter import report_preservation_derivatives  # noqa: F401
from AIPscan.Reporter import report_storage_locations  # noqa: F401
from AIPscan.Reporter import reporter  # noqa: F401
from AIPscan.Reporter import request_params  # noqa: F401
from AIPscan.Reporter import sort_puids  # noqa: F401
from AIPscan.Reporter.database_helpers import get_possible_storage_locations
from AIPscan.Reporter.helpers import calculate_paging_window
from AIPscan.Reporter.helpers import remove_dict_none_values


def _get_default_storage_service():
    """Return the default StorageService if present, otherwise the first one or None."""
    storage_service = StorageService.query.filter_by(default=True).first()
    if storage_service is None:
        storage_service = StorageService.query.first()
    return storage_service


def _get_storage_service(storage_service_id):
    """Get Storage Service from specified ID.

    If the requested Storage Service ID is invalid, return (in order of
    preference) one of the following:
    - First default Storage Service
    - First Storage Service
    - None

    :param storage_service_id: Storage Service ID

    :returns: StorageService object or None
    """
    if not storage_service_id:
        return _get_default_storage_service()

    storage_service = db.session.get(StorageService, storage_service_id)
    if storage_service is None:
        storage_service = _get_default_storage_service()

    return storage_service


def _get_storage_location(storage_location_id):
    """Return Storage Location or None."""
    if not storage_location_id:
        return None
    return db.session.get(StorageLocation, storage_location_id)


def storage_locations_with_aips(storage_locations):
    """Return list of Storage Locations filtered to those with AIPs."""
    return [loc for loc in storage_locations if loc.aips]


def get_aip_pager(page, per_page, storage_service, **kwargs):
    storage_service_id = None
    if storage_service is not None:
        storage_service_id = storage_service.id

    storage_location_id = None
    if kwargs.get("storage_location", None):
        storage_location_id = kwargs["storage_location"].id

    query = kwargs.get("query", None)

    aips = AIP.query

    # Filter by storage service
    aips = aips.filter_by(storage_service_id=storage_service_id)

    # Optionally filter by storage location
    if storage_location_id:
        aips = aips.filter_by(storage_location_id=storage_location_id)

    # Optionally filter by text search
    if query is not None:
        aips = aips.filter(
            AIP.uuid.like(f"%{query}%")
            | AIP.transfer_name.like(f"%{query}%")
            | AIP.create_date.like(f"%{query}%")
        )

    pager = aips.paginate(page=page, per_page=per_page, error_out=False)

    return pager


def get_file_pager(page, per_page, aip):
    try:
        page = int(page)
    except ValueError:
        page = 1

    return File.query.filter_by(aip_id=aip.id, file_type=FileType.original).paginate(
        page=page, per_page=per_page, error_out=False
    )


@reporter.route("/aips/", methods=["GET"])
def view_aips():
    """Overview of AIPs in given Storage Service and Location."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_service = _get_storage_service(storage_service_id)

    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    try:
        storage_location = _get_storage_location(storage_location_id)
    except Exception:
        current_app.logger.exception(
            f"Failed to get storage location id={storage_location_id}"
        )

    query = request.args.get("query", "")

    try:
        page = int(request.args.get(request_params.PAGE, default=1))
    except ValueError:
        page = 1

    pager = get_aip_pager(
        page, 10, storage_service, storage_location=storage_location, query=query
    )

    first_item, last_item = calculate_paging_window(pager)

    state_query_params = {
        request_params.STORAGE_SERVICE_ID: storage_service_id,
        request_params.STORAGE_LOCATION_ID: storage_location_id,
    }

    return render_template(
        "aips.html",
        storage_services=StorageService.query.all(),
        storage_service=storage_service,
        storage_locations=storage_locations_with_aips(
            get_possible_storage_locations(storage_service)
        ),
        storage_location=storage_location,
        pager=pager,
        first_item=first_item,
        last_item=last_item,
        state_query_params=remove_dict_none_values(state_query_params),
    )


# Picturae TODO: Does this work with AIP UUID as well?
@reporter.route("/aip/<aip_id>", methods=["GET"])
def view_aip(aip_id):
    """Detailed view of specific AIP."""
    aip = db.session.get(AIP, aip_id)

    if aip is None:
        abort(404)

    fetch_job = db.session.get(FetchJob, aip.fetch_job_id)
    storage_service = db.session.get(StorageService, fetch_job.storage_service_id)
    storage_location = db.session.get(StorageLocation, aip.storage_location_id)
    aips_count = AIP.query.filter_by(storage_service_id=storage_service.id).count()
    original_file_count = aip.original_file_count
    preservation_file_count = aip.preservation_file_count
    origin_pipeline = db.session.get(Pipeline, aip.origin_pipeline_id)

    originals = []

    page = request.args.get(request_params.PAGE, default="1")
    pager = get_file_pager(page, 10, aip)

    first_item, last_item = calculate_paging_window(pager)

    for file_ in pager.items:
        original = {}
        original["id"] = file_.id
        original["name"] = file_.name
        original["uuid"] = file_.uuid
        original["size"] = file_.size
        original["date_created"] = file_.date_created.strftime("%Y-%m-%d")
        original["puid"] = file_.puid
        original["file_format"] = file_.file_format
        original["format_version"] = file_.format_version
        preservation_file = File.query.filter_by(
            file_type=FileType.preservation, original_file_id=file_.id
        ).first()
        if preservation_file is not None:
            original["preservation_file_id"] = preservation_file.id
        originals.append(original)

    return render_template(
        "aip.html",
        aip=aip,
        first_item=first_item,
        last_iten=last_item,
        pager=pager,
        storage_service=storage_service,
        storage_location=storage_location,
        aips_count=aips_count,
        originals=originals,
        original_file_count=original_file_count,
        preservation_file_count=preservation_file_count,
        origin_pipeline=origin_pipeline,
    )


@reporter.route("/file/<file_id>", methods=["GET"])
def view_file(file_id):
    """File page displays Object and Event metadata for file"""
    file_ = db.session.get(File, file_id)

    if file_ is None:
        abort(404)

    aip = db.session.get(AIP, file_.aip_id)
    events = Event.query.filter_by(file_id=file_id).all()
    preservation_file = File.query.filter_by(
        file_type=FileType.preservation, original_file_id=file_.id
    ).first()

    original_filename = None
    if file_.original_file_id is not None:
        original_file = db.session.get(File, file_.original_file_id)
        original_filename = original_file.name

    return render_template(
        "file.html",
        file_=file_,
        premisxml=file_.get_premis_xml_lines(),
        aip=aip,
        events=events,
        preservation_file=preservation_file,
        original_filename=original_filename,
    )


@reporter.route("/reports/", methods=["GET"])
def reports():
    """Reports page lists available built-in reports

    Storage Service ID can be specified as an optional parameter. If
    one is not specified, use the default Storage Service.
    """
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_service = _get_storage_service(storage_service_id)

    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    storage_location = _get_storage_location(storage_location_id)

    original_file_formats = []
    preservation_file_formats = []
    original_puids = []
    preservation_puids = []

    # Filter dropdowns by Storage Location if one is selected. If a location is
    # not selected, populate dropdowns from Storage Service instead.
    if storage_location:
        # Original file formats and PUIDs should be present for any Storage
        # Loocation with ingested files.
        try:
            original_file_formats = storage_location.unique_original_file_formats
            original_puids = sort_puids(storage_location.unique_original_puids)
        except AttributeError:
            pass

        # Preservation file formats and PUIDs may be present depending on
        # a number of factors such as processing configuration choices,
        # normalization rules, and use of manual normalization.
        try:
            preservation_file_formats = (
                storage_location.unique_preservation_file_formats
            )
        except AttributeError:
            pass
        try:
            preservation_puids = sort_puids(storage_location.unique_preservation_puids)
        except AttributeError:
            pass

    else:
        # Original file formats and PUIDs should be present for any Storage
        # Service with ingested files.
        try:
            original_file_formats = storage_service.unique_original_file_formats
            original_puids = sort_puids(storage_service.unique_original_puids)
        except AttributeError:
            pass

        # Preservation file formats and PUIDs may be present depending on
        # a number of factors such as processing configuration choices,
        # normalization rules, and use of manual normalization.
        try:
            preservation_file_formats = storage_service.unique_preservation_file_formats
        except AttributeError:
            pass
        try:
            preservation_puids = sort_puids(storage_service.unique_preservation_puids)
        except AttributeError:
            pass

    if "start_date" not in session:
        earliest_aip_created = storage_service.earliest_aip_created
        session["start_date"] = earliest_aip_created.strftime("%Y-%m-%d")

    if "end_date" not in session:
        now = datetime.now()
        session["end_date"] = str(datetime(now.year, now.month, now.day))[:-9]

    return render_template(
        "reports.html",
        storage_service=storage_service,
        storage_services=StorageService.query.all(),
        storage_location=storage_location,
        storage_locations=storage_locations_with_aips(
            get_possible_storage_locations(storage_service)
        ),
        original_file_formats=original_file_formats,
        preservation_file_formats=preservation_file_formats,
        original_puids=original_puids,
        preservation_puids=preservation_puids,
        start_date=session["start_date"],
        end_date=session["end_date"],
    )


@reporter.route("/update_dates/", methods=["POST"])
def update_dates():
    if request.json and request.json.get("start_date"):
        req = request.get_json()
        session["start_date"] = request.json.get("start_date")
        return make_response(jsonify(req), 200)

    if request.json and request.json.get("end_date"):
        req = request.get_json()
        session["end_date"] = request.json.get("end_date")
        return make_response(jsonify(req), 200)


@reporter.route("/download_mets/<aip_id>", methods=["GET"])
def download_mets(aip_id):
    aip = db.session.get(AIP, aip_id)
    storage_service = db.session.get(StorageService, aip.storage_service_id)

    mets_response = requests.get(
        get_mets_url(
            storage_service,
            aip.uuid,
            f"{aip.transfer_name}-{aip.uuid}/data/METS.{aip.uuid}.xml",
        )
    )

    headers = {
        "Content-Disposition": f'attachment; filename="METS-{aip.uuid}.xml"',
        "Content-length": len(mets_response.content),
    }
    return Response(mets_response.content, mets_response.status_code, headers)



"""Test the efficacy of the agents_transfers function which supplies
the ingest log report and agents-transfers API endpoint.
"""

import pytest

from AIPscan.conftest import AIP_CREATION_TIME
from AIPscan.conftest import INGEST_EVENT_CREATION_TIME
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_datetime_bound

DAY_BEFORE_AIP_CREATION = parse_datetime_bound("2020-12-01")
DAY_OF_AIP_CREATION = parse_datetime_bound("2020-12-03", upper=True)
DAY_WELL_PAST_AIP_CREATION = parse_datetime_bound("2022-04-12")


@pytest.mark.parametrize(
    "storage_id, storage_name, location_id, location_name, number_of_ingests, aip_uuid, aip_name, user_name, ingest_date, aip_creation_date, start_date, end_date",
    [
        # Simple transfer with limited amount of data from the DB.
        (
            1,
            "test storage service",
            1,
            "test storage location",
            1,
            "111111111111-1111-1111-11111111",
            "Test AIP",
            "user one",
            INGEST_EVENT_CREATION_TIME,
            AIP_CREATION_TIME,
            DAY_BEFORE_AIP_CREATION,
            DAY_OF_AIP_CREATION,
        ),
        # Request for a non-existent storage service.
        (30, None, 35, None, 0, "", "", "", "", "", "", ""),
        # Transfer that occured outside of specified date range.
        (
            1,
            "test storage service",
            1,
            "test storage location",
            0,
            "",
            "",
            "",
            "",
            "",
            DAY_WELL_PAST_AIP_CREATION,
            DAY_WELL_PAST_AIP_CREATION,
        ),
    ],
)
def test_agents_transfers(
    app_with_populated_files,
    storage_id,
    storage_name,
    location_id,
    location_name,
    number_of_ingests,
    aip_uuid,
    aip_name,
    user_name,
    ingest_date,
    aip_creation_date,
    start_date,
    end_date,
):
    """Test that structure of agents-transfers matches expectations."""
    report = report_data.agents_transfers(
        storage_service_id=storage_id,
        start_date=start_date,
        end_date=end_date,
        storage_location_id=location_id,
    )

    assert report[fields.FIELD_STORAGE_NAME] == storage_name
    assert report[fields.FIELD_STORAGE_LOCATION] == location_name
    assert len(report[fields.FIELD_INGESTS]) == number_of_ingests

    if number_of_ingests < 1:
        return

    ingests = report["Ingests"]
    assert ingests[0]["AIPUUID"] == aip_uuid
    assert ingests[0]["AIPName"] == aip_name
    assert ingests[0]["User"] == "user one"
    assert ingests[0]["IngestStartDate"] == str(ingest_date)
    assert ingests[0]["IngestFinishDate"] == str(aip_creation_date)


@pytest.mark.parametrize(
    "storage_id, storage_name, location_id, location_name, number_of_ingests, start_date, end_date",
    [
        # Simple transfer with limited amount of data from the DB.
        (
            1,
            "test storage service",
            1,
            "test storage location",
            0,
            DAY_BEFORE_AIP_CREATION,
            DAY_OF_AIP_CREATION,
        )
    ],
)
def test_agents_transfers_no_ingestion_event(
    app_with_populated_files_no_ingestion_event,
    storage_id,
    storage_name,
    location_id,
    location_name,
    number_of_ingests,
    start_date,
    end_date,
):
    """Test that lack of ingestion event doesn't throw error."""
    report = report_data.agents_transfers(
        storage_service_id=storage_id,
        start_date=start_date,
        end_date=end_date,
        storage_location_id=location_id,
    )

    assert report[fields.FIELD_STORAGE_NAME] == storage_name
    assert report[fields.FIELD_STORAGE_LOCATION] == location_name
    assert len(report[fields.FIELD_INGESTS]) == number_of_ingests



"""Collects a number of reusable components of tasks.py. Also ensures
the module remains clean and easy to refactor over time.
"""

import json
import os
from datetime import datetime

from dateutil.parser import ParserError
from dateutil.parser import parse

from AIPscan.Aggregator.downloads import get_download_root
from AIPscan.Aggregator.types import StorageServicePackage


def format_api_url_with_limit_offset(storage_service):
    """Format the API URL here to make sure it is as correct as
    possible.
    """
    base_url = storage_service.url.rstrip("/")

    request_url_without_api_key = f"{base_url}/api/v2/file/?limit={storage_service.download_limit}&offset={storage_service.download_offset}"
    request_url = f"{request_url_without_api_key}&username={storage_service.user_name}&api_key={storage_service.api_key}"
    return base_url, request_url_without_api_key, request_url


def get_packages_directory(timestamp):
    """Create a path which we will use to store packages downloaded from
    the storage service plus other metadata.
    """
    return os.path.join(get_download_root(), timestamp, "packages")


def parse_package_list_file(filepath, logger=None, remove_after_parsing=False):
    with open(filepath) as packages_json:
        package_list = json.load(packages_json)
    try:
        if remove_after_parsing:
            os.remove(filepath)
    except OSError as err:
        if logger:
            logger.warning(f"Unable to delete package JSON file: {err}")

    return package_list


def process_package_object(package_obj):
    """Process a package object as retrieve from the storage service
    and return a StorageServicePackage type to the caller for further
    analysis.
    """
    package = StorageServicePackage()

    STATUS = "status"
    TYPE = "package_type"
    REPL = "replicated_package"
    CURRENT_LOCATION = "current_location"
    CURRENT_PATH = "current_path"
    ORIGIN_PIPELINE = "origin_pipeline"
    UUID = "uuid"

    PKG_AIP = "AIP"
    PKG_SIP = "transfer"
    PKG_DIP = "DIP"
    PKG_DEL = "DELETED"

    # Accumulate state. The package object should be able to evaluate
    # itself accordingly.
    if package_obj[TYPE] == PKG_AIP:
        package.aip = True
    if package_obj.get(TYPE) == PKG_DIP:
        package.dip = True
    if package_obj.get(TYPE) == PKG_SIP:
        package.sip = True
    if package_obj.get(STATUS) == PKG_DEL:
        package.deleted = True
    if package_obj.get(REPL) is not None:
        package.replica = True

    package.uuid = package_obj.get(UUID)
    package.current_location = package_obj.get(CURRENT_LOCATION)
    package.current_path = package_obj.get(CURRENT_PATH)
    package.origin_pipeline = package_obj.get(ORIGIN_PIPELINE)
    package.size = package_obj.get("size")

    return package


def _tz_neutral_date(date):
    """Convert inconsistent dates consistently. Dates are round-tripped
    back to a Python datetime object as anticipated by the database.
    Where a date is unknown or can't be parsed, we return the UNIX EPOCH
    in lieu of another sensible value.
    """
    date_time_pattern = "%Y-%m-%dT%H:%M:%S"
    EPOCH = datetime.strptime("1970-01-01T00:00:01", date_time_pattern)
    try:
        date = parse(date)
        date = date.strftime(date_time_pattern)
        date = datetime.strptime(date, date_time_pattern)
    except ParserError:
        date = EPOCH
    return date


def get_mets_url(storage_service, package_uuid, path_to_mets):
    """Construct a URL from which we can download the METS files that
    we are interested in.
    """
    mets_url = "{}/api/v2/file/{}/extract_file/?relative_path_to_file={}&username={}&api_key={}".format(
        storage_service.url.rstrip("/"),
        package_uuid,
        path_to_mets,
        storage_service.user_name,
        storage_service.api_key,
    )
    return mets_url


def get_storage_service_api_url(storage_service, api_path):
    """Return URL to fetch location infofrom Storage Service."""
    base_url = storage_service.url.rstrip("/")
    request_url_without_api_key = f"{base_url}{api_path}".rstrip("/")

    request_url = f"{request_url_without_api_key}?username={storage_service.user_name}&api_key={storage_service.api_key}"
    return request_url, request_url_without_api_key


def create_numbered_subdirs(timestamp, package_list_number):
    """Check for the existence and create a container folder for our
    METS files as required.
    """
    METS_FOLDER = "mets"

    # Create a package list numbered subdirectory if it doesn't exist.
    numbered_subdir = os.path.join(
        get_download_root(), timestamp, METS_FOLDER, str(package_list_number)
    )
    if not os.path.exists(numbered_subdir):
        os.makedirs(numbered_subdir)

    return numbered_subdir


def summarize_fetch_job_results(fetch_job):
    return f"aips: '{fetch_job.total_aips}'; sips: '{fetch_job.total_sips}'; dips: '{fetch_job.total_dips}'; deleted: '{fetch_job.total_deleted_aips}'; replicated: '{fetch_job.total_replicas}'"



"""Initial schema migration.

Revision ID: c605f2284613
Revises:
Create Date: 2025-10-20 00:00:00.000000

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# Revision identifiers are used by Alembic.
revision = "c605f2284613"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### Commands auto-generated by Alembic. Please adjust. ###
    op.create_table(
        "pipeline",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("origin_pipeline", sa.String(length=255), nullable=True),
        sa.Column("dashboard_url", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_pipeline_origin_pipeline"), ["origin_pipeline"], unique=True
        )

    op.create_table(
        "storage_service",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("url", sa.String(length=255), nullable=True),
        sa.Column("user_name", sa.String(length=255), nullable=True),
        sa.Column("api_key", sa.String(length=255), nullable=True),
        sa.Column("download_limit", sa.Integer(), nullable=True),
        sa.Column("download_offset", sa.Integer(), nullable=True),
        sa.Column("default", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("storage_service", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_storage_service_name"), ["name"], unique=True
        )

    op.create_table(
        "agent",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("linking_type_value", sa.String(length=255), nullable=True),
        sa.Column("agent_type", sa.String(length=255), nullable=True),
        sa.Column("agent_value", sa.String(length=255), nullable=True),
        sa.Column("storage_service_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["storage_service_id"],
            ["storage_service.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("agent", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_agent_agent_type"), ["agent_type"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_agent_agent_value"), ["agent_value"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_agent_linking_type_value"),
            ["linking_type_value"],
            unique=False,
        )

    op.create_table(
        "fetch_job",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("total_packages", sa.Integer(), nullable=True),
        sa.Column("total_aips", sa.Integer(), nullable=True),
        sa.Column("total_dips", sa.Integer(), nullable=True),
        sa.Column("total_sips", sa.Integer(), nullable=True),
        sa.Column("total_replicas", sa.Integer(), nullable=True),
        sa.Column("total_deleted_aips", sa.Integer(), nullable=True),
        sa.Column("download_start", sa.DateTime(), nullable=True),
        sa.Column("download_end", sa.DateTime(), nullable=True),
        sa.Column("download_directory", sa.String(length=255), nullable=True),
        sa.Column("storage_service_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["storage_service_id"],
            ["storage_service.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "storage_location",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("current_location", sa.String(length=255), nullable=True),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("storage_service_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["storage_service_id"],
            ["storage_service.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("storage_location", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_storage_location_current_location"),
            ["current_location"],
            unique=True,
        )

    op.create_table(
        "aip",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=255), nullable=True),
        sa.Column("transfer_name", sa.String(length=255), nullable=True),
        sa.Column("create_date", sa.DateTime(), nullable=True),
        sa.Column("mets_sha256", sa.String(length=64), nullable=True),
        sa.Column("size", sa.BigInteger(), nullable=True),
        sa.Column("storage_service_id", sa.Integer(), nullable=False),
        sa.Column("storage_location_id", sa.Integer(), nullable=False),
        sa.Column("fetch_job_id", sa.Integer(), nullable=False),
        sa.Column("origin_pipeline_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["fetch_job_id"],
            ["fetch_job.id"],
        ),
        sa.ForeignKeyConstraint(
            ["origin_pipeline_id"],
            ["pipeline.id"],
        ),
        sa.ForeignKeyConstraint(
            ["storage_location_id"],
            ["storage_location.id"],
        ),
        sa.ForeignKeyConstraint(
            ["storage_service_id"],
            ["storage_service.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("aip", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_aip_create_date"), ["create_date"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_aip_uuid"), ["uuid"], unique=False)

    op.create_table(
        "index_tasks",
        sa.Column("index_task_id", sa.String(length=36), nullable=False),
        sa.Column("fetch_job_id", sa.Integer(), nullable=False),
        sa.Column("indexing_start", sa.DateTime(), nullable=True),
        sa.Column("indexing_progress", sa.String(length=255), nullable=True),
        sa.Column("indexing_end", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["fetch_job_id"],
            ["fetch_job.id"],
        ),
        sa.PrimaryKeyConstraint("index_task_id"),
    )
    op.create_table(
        "file",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("filepath", sa.Text(), nullable=True),
        sa.Column("uuid", sa.String(length=255), nullable=True),
        sa.Column(
            "file_type",
            sa.Enum("original", "preservation", name="filetype"),
            nullable=True,
        ),
        sa.Column("size", sa.BigInteger(), nullable=True),
        sa.Column("date_created", sa.DateTime(), nullable=True),
        sa.Column("puid", sa.String(length=255), nullable=True),
        sa.Column("file_format", sa.String(length=255), nullable=True),
        sa.Column("format_version", sa.String(length=255), nullable=True),
        sa.Column("checksum_type", sa.String(length=255), nullable=True),
        sa.Column("checksum_value", sa.String(length=255), nullable=True),
        sa.Column("premis_object", mysql.LONGTEXT(), nullable=True),
        sa.Column("original_file_id", sa.Integer(), nullable=True),
        sa.Column("aip_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["aip_id"],
            ["aip.id"],
        ),
        sa.ForeignKeyConstraint(
            ["original_file_id"],
            ["file.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("file", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_file_checksum_value"), ["checksum_value"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_file_date_created"), ["date_created"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_file_name"), ["name"], unique=False)
        batch_op.create_index(batch_op.f("ix_file_puid"), ["puid"], unique=False)
        batch_op.create_index(batch_op.f("ix_file_uuid"), ["uuid"], unique=False)

    op.create_table(
        "event",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=255), nullable=True),
        sa.Column("uuid", sa.String(length=255), nullable=True),
        sa.Column("date", sa.DateTime(), nullable=True),
        sa.Column("detail", sa.String(length=255), nullable=True),
        sa.Column("outcome", sa.String(length=255), nullable=True),
        sa.Column("outcome_detail", mysql.LONGTEXT(), nullable=True),
        sa.Column("file_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["file_id"],
            ["file.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("event", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_event_type"), ["type"], unique=False)
        batch_op.create_index(batch_op.f("ix_event_uuid"), ["uuid"], unique=False)

    op.create_table(
        "event_agents",
        sa.Column("event_id", sa.Integer(), nullable=True),
        sa.Column("agent_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["agent.id"],
        ),
        sa.ForeignKeyConstraint(
            ["event_id"],
            ["event.id"],
        ),
    )
    op.create_table(
        "get_mets_tasks",
        sa.Column("get_mets_task_id", sa.String(length=36), nullable=False),
        sa.Column("workflow_coordinator_id", sa.String(length=36), nullable=True),
        sa.Column("package_uuid", sa.String(length=36), nullable=True),
        sa.Column("status", sa.String(length=36), nullable=True),
        sa.PrimaryKeyConstraint("get_mets_task_id"),
    )
    op.create_table(
        "package_tasks",
        sa.Column("package_task_id", sa.String(length=36), nullable=False),
        sa.Column("workflow_coordinator_id", sa.String(length=36), nullable=True),
        sa.PrimaryKeyConstraint("package_task_id"),
    )
    # ### End Alembic commands. ###


def downgrade():
    # ### Commands auto-generated by Alembic. Please adjust. ###
    op.drop_table("package_tasks")
    op.drop_table("get_mets_tasks")
    op.drop_table("event_agents")
    with op.batch_alter_table("event", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_event_uuid"))
        batch_op.drop_index(batch_op.f("ix_event_type"))

    op.drop_table("event")
    with op.batch_alter_table("file", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_file_uuid"))
        batch_op.drop_index(batch_op.f("ix_file_puid"))
        batch_op.drop_index(batch_op.f("ix_file_name"))
        batch_op.drop_index(batch_op.f("ix_file_date_created"))
        batch_op.drop_index(batch_op.f("ix_file_checksum_value"))

    op.drop_table("file")
    op.drop_table("index_tasks")
    with op.batch_alter_table("aip", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_aip_uuid"))
        batch_op.drop_index(batch_op.f("ix_aip_create_date"))

    op.drop_table("aip")
    with op.batch_alter_table("storage_location", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_storage_location_current_location"))

    op.drop_table("storage_location")
    op.drop_table("fetch_job")
    with op.batch_alter_table("agent", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_agent_linking_type_value"))
        batch_op.drop_index(batch_op.f("ix_agent_agent_value"))
        batch_op.drop_index(batch_op.f("ix_agent_agent_type"))

    op.drop_table("agent")
    with op.batch_alter_table("storage_service", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_storage_service_name"))

    op.drop_table("storage_service")
    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_pipeline_origin_pipeline"))

    op.drop_table("pipeline")
    # ### End Alembic commands. ###



from flask import current_app

EXPECTED_CSV_CONTENTS = b"PUID,Format,Version,Count,Size,Size (bytes)\r\nfmt/44,JPEG,1.02,1,2.0 kB,2000\r\nfmt/43,JPEG,1.01,1,1.0 kB,1000\r\nfmt/468,ISO Disk Image File,,1,0 Bytes,0\r\n"


def test_format_versions_count(app_with_populated_format_versions):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/report_format_versions_count/?amss_id=1")
        assert response.status_code == 200


def test_format_versions_count_csv(app_with_populated_format_versions):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            "/reporter/report_format_versions_count/?amss_id=1&csv=True"
        )
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=format_versions.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS



from AIPscan import create_app


def main():
    """Entry point for running the Flask application using the built-in server.

    This is only for development purposes, use Gunicorn or similar for production.
    """
    app = create_app()
    app.run(host="0.0.0.0", port=5000)


if __name__ == "__main__":
    main()



from flask import render_template
from flask import request

from AIPscan import typesense_helpers as ts_helpers
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

TABLE_HEADERS = [
    fields.FIELD_FILENAME,
    fields.FIELD_SIZE,
    fields.FIELD_FORMAT,
    fields.FIELD_PUID,
    fields.FIELD_FILE_TYPE,
    fields.FIELD_AIP_NAME,
]

CSV_HEADERS = [
    fields.FIELD_UUID,
    fields.FIELD_FILENAME,
    fields.FIELD_SIZE,
    fields.FIELD_FILE_TYPE,
    fields.FIELD_FORMAT,
    fields.FIELD_VERSION,
    fields.FIELD_PUID,
    fields.FIELD_AIP_NAME,
    fields.FIELD_AIP_UUID,
]


@reporter.route("/largest_files/", methods=["GET"])
def largest_files():
    """Return largest files."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    file_type = request.args.get(request_params.FILE_TYPE)
    limit = 20
    try:
        limit = int(request.args.get(request_params.LIMIT, 20))
    except ValueError:
        pass
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    if ts_helpers.typesense_enabled():
        file_data = report_data_typesense.largest_files(
            storage_service_id,
            start_date,
            end_date,
            storage_location_id,
            file_type,
            limit,
        )
    else:
        file_data = report_data.largest_files(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
            file_type=file_type,
            limit=limit,
        )

    if csv:
        headers = translate_headers(CSV_HEADERS, True)

        filename = "largest_files.csv"
        csv_data = format_size_for_csv(file_data[fields.FIELD_FILES])
        return download_csv(headers, csv_data, filename)

    headers = translate_headers(TABLE_HEADERS)

    return render_template(
        "report_largest_files.html",
        storage_service_id=storage_service_id,
        storage_service_name=file_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_id=storage_location_id,
        storage_location_description=file_data.get(fields.FIELD_STORAGE_LOCATION),
        columns=headers,
        files=file_data[fields.FIELD_FILES],
        file_type=file_type,
        limit=limit,
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )
